###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import uuid
import json
import binascii
import pkg_resources
from pprint import pformat

import requests
import txaio

from twisted.internet.defer import DeferredList, inlineCallbacks, Deferred

import zlmdb

from autobahn import wamp
from autobahn import xbr
from autobahn.xbr import make_w3
from autobahn.wamp.types import ComponentConfig, RegisterOptions
from autobahn.wamp.request import Registration
from autobahn.twisted.wamp import ApplicationSession

from crossbar._util import hltype, hlid, hl, hlval
from crossbar.common import checkconfig

from crossbar._util import merge_config
from crossbar.node.node import Node
from crossbar.edge.node import node
from crossbar.master.node.roles import BUILTIN_ROLES
from crossbar.master.mrealm.mrealm import ManagementRealm
from cfxdb.globalschema import GlobalSchema

__all__ = ('FabricServiceNodeManager', 'FabricCenterNode')

_CFC_GLOBAL_REALM = 'crossbar.'


class License(object):
    def __init__(self, license_type):
        assert license_type in [20010]

        self._type = license_type

        if license_type == 20010:
            self._title = 'crossbar Trial'
            self._mrealms = 1
            self._routers = 2
            self._proxies = 2
            self._containers = 4
            self._guests = 4
            self._marketmakers = 1

    @property
    def type(self):
        return self._type

    @property
    def title(self):
        return self._title

    @property
    def mrealms(self):
        return self._mrealms

    @property
    def routers(self):
        return self._routers

    @property
    def containers(self):
        return self._containers

    @property
    def guests(self):
        return self._guests

    @property
    def proxies(self):
        return self._proxies

    @property
    def marketmakers(self):
        return self._marketmakers

    def marshal(self):
        obj = {
            'type': self._type,
            'title': self._title,
            'mrealms': self._mrealms,
            'routers': self._routers,
            'containers': self._containers,
            'guests': self._guests,
            'proxies': self._proxies,
            'marketmakers': self._marketmakers,
        }
        return obj

    def __str__(self):
        return '{}'.format(self.marshal())


class FabricServiceNodeManager(ApplicationSession):
    """
    CFC Node manager. This is responsible for starting and stopping
    management realms
    """

    log = txaio.make_logger()

    def __init__(self, config, reactor, personality, node):
        # base ctor
        super(ApplicationSession, self).__init__(config=config)

        self._reactor = reactor
        self.personality = personality
        self._node = node

        # database and schema
        self._db = None
        self._schema = None

        # XBR API Hub
        self._xbr = None

    async def onJoin(self, details: ComponentConfig):

        # handles to controller database and schema (already initialized/attached in node)
        #
        ready = self.config.extra['ready']
        cbdir = self.config.extra['cbdir']
        config = self.config.extra.get('database', {})

        # create database and attach tables to database slots
        #
        dbpath = config.get('path', '.db-controller')
        assert type(dbpath) == str
        dbpath = os.path.join(cbdir, dbpath)

        maxsize = config.get('maxsize', 128 * 2**20)
        assert type(maxsize) == int
        # allow maxsize 128kiB to 128GiB
        assert maxsize >= 128 * 1024 and maxsize <= 128 * 2**30

        self.log.info('{klass} starting [dbpath={dbpath}, maxsize={maxsize}]',
                      klass=self.__class__.__name__,
                      dbpath=hlid(dbpath),
                      maxsize=hlid(maxsize))

        self._db = zlmdb.Database(dbpath=dbpath, maxsize=maxsize, readonly=False, sync=True)
        self._db.__enter__()
        self._schema = GlobalSchema.attach(self._db)

        # initialize all currently existing mrealms
        try:
            await self._initialize_mrealms()
        except Exception:
            self.log.failure()
            raise

        # expose api on this object fo CFC clients
        domains = [(_CFC_GLOBAL_REALM, self.register)]
        for prefix, register in domains:
            registrations = await register(self, prefix=prefix, options=RegisterOptions(details_arg='details'))
            for reg in registrations:
                if type(reg) == Registration:
                    self.log.info('Registered CFC Global Realm "{realm}" API <{proc}>',
                                  proc=reg.procedure,
                                  realm=self._realm)
                else:
                    self.log.error('Error: <{}>'.format(reg.value.args[0]))

        # we are ready to roll
        self.log.info('{note} {klass}',
                      note=hl('Ok, master node "{}" booted and ready!'.format(self._node._node_id),
                              color='red',
                              bold=True),
                      klass=hltype(self.onJoin))

        ready.callback(self)

    def onLeave(self, details):
        pass

    async def _initialize_mrealms(self, parallel=False):

        self._router_realms = {}
        self._container_workers = {}

        with self._db.begin() as txn:
            mrealms = list(self._schema.mrealms.select(txn, return_keys=False))

        if mrealms:
            self.log.info('Initializing {cnt_mrealms} management realms (parallel={parallel}) ..',
                          cnt_mrealms=hlval(len(mrealms)),
                          parallel=hlval(parallel))
            if parallel:
                dl = []
                for mrealm in mrealms:
                    dl.append(self._initialize_mrealm(mrealm))
                if dl:
                    await DeferredList(dl)
            else:
                for mrealm in mrealms:
                    await self._initialize_mrealm(mrealm)
            self.log.info('Ok, completed initializing {cnt_mrealms} management realms.',
                          cnt_mrealms=hlval(len(mrealms)))
        else:
            self.log.info('No management realms no initialize. Done.')

    async def _initialize_mrealm(self, mrealm):
        """
        Initializes a single management realm:

        1. start a respective realm on the router worker of this master node. Currently there is only one
           router worker statically configured (named "cfrouter1").
        2. start a couple of roles on the management realm (see BUILTIN_ROLES)
        3.

        :param mrealm: The management realm to initialize.
        :type mrealm: :class:`cfxdb.mrealm.management_realm.ManagementRealm`

        :return:
        """

        self.log.debug('{klass}._initialize_mrealm(mrealm={mrealm})', klass=self.__class__.__name__, mrealm=mrealm)
        # WAMP realm name of the realm to start
        realm_name = mrealm.name

        # router and container worker for realm to start
        router_id = mrealm.cf_router_worker or 'cfrouter1'
        container_id = mrealm.cf_container_worker or 'cfcontainer1'

        self.log.info(
            hl('Initializing management realm "{}" [router_id="{}", container_id="{}"] ..'.format(
                realm_name, router_id, container_id),
               color='red',
               bold=True))

        # start the management realm and roles on the configured router worker
        if realm_name not in self._router_realms:
            realm_config = {
                'name': realm_name,
                'options': {
                    'enable_meta_api': True,

                    # FIXME: enabling this will break stopping and restarting an mrealm, as the
                    # procedures registered eg by the RouterServiceAgent session will stick around!
                    'bridge_meta_api': False,
                }
            }
            realm_id = realm_name
            await self.call('crossbar.worker.{}.start_router_realm'.format(router_id), realm_id, realm_config)

            for role_id in ['owner-role', 'backend-role', 'node-role', 'public-role']:
                await self.call('crossbar.worker.{}.start_router_realm_role'.format(router_id), realm_id, role_id,
                                BUILTIN_ROLES[role_id])

            self._router_realms[realm_name] = True
        else:
            self.log.warn('Management realm "{realm_name}" already initialized (skipped starting realm)',
                          realm_name=realm_name)

        # start the configured container worker (if not yet started)
        if container_id not in self._container_workers:
            container_options = {
                "pythonpath": [".."],
                "expose_shared": True,
                "expose_controller": True,
                # "shutdown": "shutdown-on-last-component-stopped",
                "shutdown": "shutdown-manual",
                # "restart": "restart-on-failed"
                "restart": "restart-always"
            }
            await self.call('crossbar.start_worker', container_id, 'container', container_options)

            self._container_workers[container_id] = {}
        else:
            self.log.warn(
                'Management realm "{realm_name}" backend container "{container_id}" already running (skipped starting container)',
                realm_name=realm_name,
                container_id=container_id)

        # start the management realm backend in the configured container worker
        component_id = realm_name
        if component_id not in self._container_workers[container_id]:
            mrealm_backend_extra = {
                'mrealm': str(mrealm.oid),
                'database': {
                    # the mrealm database path contains the mrealm UUID
                    'dbfile': os.path.join(self.config.extra['cbdir'], '.db-mrealm-{}'.format(mrealm.oid)),

                    # hard-code max mrealm database size to 2GB
                    # https://github.com/crossbario/crossbar/issues/235
                    'maxsize': 2**30 * 2,
                },
                'controller-database': {
                    # forward controller database parameters, so that the mrealm backend can _also_ open
                    # the controller database (read-only)
                    'dbfile': self._db.dbpath,
                    'maxsize': self._db.maxsize,
                },
            }
            component_config = {
                'type': 'class',
                'classname': 'crossbar.master.mrealm.MrealmController',
                'realm': realm_name,
                'transport': {
                    # we connect back to the master router over UDS/RawSocket/CBOR
                    'type': 'rawsocket',
                    'endpoint': {
                        'type': 'unix',
                        'path': 'sock1'
                    },
                    'serializer': 'cbor',
                },
                'extra': mrealm_backend_extra
            }
            await self.call('crossbar.worker.{}.start_component'.format(container_id), component_id, component_config)

            self._container_workers[container_id][component_id] = True
        else:
            self.log.warn(
                'Management realm "{realm_name}" backend component "{component_id}" already running (skipped starting container component)',
                realm_name=realm_name,
                component_id=component_id)

        self.log.info('{note} {func}',
                      note=hl('Ok, management realm "{}" initialized!'.format(realm_name), color='red', bold=True),
                      func=hltype(self._initialize_mrealm))

    @wamp.register(None)
    async def activate_realm(self, mrealm_obj, details=None):
        mrealm = ManagementRealm.parse(mrealm_obj)
        self.log.info('Calling initialize_realm with ({})'.format(mrealm.name))
        await self._initialize_mrealm(mrealm)

    @wamp.register(None)
    async def deactivate_realm(self, mrealm_obj, details=None):
        mrealm = ManagementRealm.parse(mrealm_obj)

        realm_name = mrealm.name
        router_id = 'cfrouter1'
        container_id = 'cfcontainer1'

        if realm_name in self._router_realms:
            self.log.info('Deactivate realm ({}) - start'.format(realm_name))

            # stop the management component
            if True:
                await self.call('crossbar.worker.{}.stop_component'.format(container_id), realm_name)
                del self._container_workers[container_id][realm_name]

            # stop the router roles
            if True:
                for role_id in ['owner-role', 'backend-role', 'node-role', 'public-role']:
                    await self.call('crossbar.worker.{}.stop_router_realm_role'.format(router_id), realm_name, role_id)

            #  stop the realm
            if True:
                await self.call('crossbar.worker.{}.stop_router_realm'.format(router_id), realm_name)
                del self._router_realms[realm_name]

            self.log.info('Management realm "{realm_name}" deactivated (complete)', realm_name=realm_name)
        else:
            self.log.warn('Management realm "{realm_name}" not active (skipped deactivation)', realm_name=realm_name)


class FabricCenterNode(node.FabricNode):
    """
    Crossbar.io Master node personality.
    """
    DEFAULT_CONFIG_PATH = 'master/node/config.json'
    """
    Builtin default master node configuration.
    """
    def __init__(self, personality, cbdir=None, reactor=None, native_workers=None, options=None):
        node.FabricNode.__init__(self, personality, cbdir, reactor, native_workers, options)

        # master node embedded DB
        self._db = None
        self._schema = None

        # Web3 blockchain connection (if configured)
        self._w3 = None

        # the node license under which this master node will be running (instance of :class:License)
        self._license = None

    def load_config(self, configfile=None):
        """
        Load the node configuration of CFC itself. The base node configuration
        is built into CFC, but can be overridden (partially) and extended
        by providing a regular, local node configuration file.

        When such a file exists:

        - and it contains a controller section, these take precedence over
        the builtin values.
        - and it contains a workers section (a list of workers), these are
        _added_ to the builtin workers.
        """
        config_source = Node.CONFIG_SOURCE_EMPTY

        # 1/4: load builtin master node configuration as default
        #
        config_path = pkg_resources.resource_filename('crossbar', self.DEFAULT_CONFIG_PATH)
        with open(config_path) as f:
            config = json.load(f)

        # no need to check the builtin config (at run-time)
        # self.personality.check_config(self.personality, config)

        # remember config source
        config_source += Node.CONFIG_SOURCE_DEFAULT
        self.log.info('Built-in master node configuration loaded')

        # 2/4: allow extending/overriding the node configuration
        # with a local master node configuration file (eg for TLS setup)
        if configfile:
            config_path = os.path.abspath(os.path.join(self._cbdir, configfile))
            with open(config_path) as f:
                custom_config = json.load(f)

            # as an overriding config file does not need to be a fully valid config in itself,
            # do _not_ check it ..
            # self.personality.check_config(self.personality, custom_config)

            # .. instead, we merge the config from the local file into the already
            # loaded default config (see above)
            config = merge_config(config, custom_config)

            # remember config source
            config_source += Node.CONFIG_SOURCE_LOCALFILE
            self.log.info('Local master node configuration merged from "{config_path}"',
                          config_path=hlval(config_path))

        # 3/4: allow extending/overriding the node configuration
        # with a remote master node configuration file from blockchain/IPFS
        if config.get('controller', {}).get('blockchain', None):
            blockchain_config = config['controller']['blockchain']
            gateway_config = blockchain_config['gateway']

            # setup Web3 connection from blockchain gateway configuration
            self._w3 = make_w3(gateway_config)

            # configure new Web3 connection as provider for XBR
            xbr.setProvider(self._w3)

            if self._w3.isConnected():
                self.log.info('Connected to Ethereum network {network} at gateway "{gateway_url}"',
                              network=self._w3.net.version,
                              gateway_url=gateway_config['http'])

                # try to find node by node public key on-chain
                xbr_node_id = xbr.xbrnetwork.functions.getNodeByKey(self.key.public_key()).call()
                if xbr_node_id != b'\x00' * 16:
                    assert (len(xbr_node_id) == 16)

                    # get node domain, type, configuration and license as stored on-chain
                    xbr_node_domain = xbr.xbrnetwork.functions.getNodeDomain(xbr_node_id).call()
                    xbr_node_type = xbr.xbrnetwork.functions.getNodeType(xbr_node_id).call()
                    xbr_node_config = xbr.xbrnetwork.functions.getNodeConfig(xbr_node_id).call()

                    # FIXME: for testing use hard-coded "trial license" (no. 20010)
                    xbr_node_license = 20010
                    # xbr_node_license = xbr.xbrnetwork.functions.getNodeLicense(xbr_node_id).call()

                    self.log.info(
                        'Node is configured for XBR (xbr_node_id="{xbr_node_id}", xbr_node_domain="{xbr_node_domain}", xbr_node_type="{xbr_node_type}", xbr_node_license="{xbr_node_license}", xbr_node_config="{xbr_node_config}")',
                        xbr_node_id=hlid('0x' + binascii.b2a_hex(xbr_node_id).decode()),
                        xbr_node_domain=hlid('0x' + binascii.b2a_hex(xbr_node_domain).decode()),
                        xbr_node_type=hlid(xbr_node_type),
                        xbr_node_config=hlid(xbr_node_config),
                        xbr_node_license=hlid(xbr_node_license))

                    self._license = License(xbr_node_license)
                    self.log.info('Node is registered in the XBR network with license type={license}',
                                  license=self._license.type)

                    # if a (hash of a) configuration is set on-chain for the master node, fetch the
                    # configuration content for the hash from IPFS
                    if xbr_node_config:
                        if 'IPFS_GATEWAY_URL' in os.environ:
                            ipfs_gateway_url = os.environ['IPFS_GATEWAY_URL']
                            self.log.info(
                                'Using explicit IPFS gateway URL {ipfs_gateway_url} from environment variable IPFS_GATEWAY_URL',
                                ipfs_gateway_url=hlid(ipfs_gateway_url))
                        else:
                            ipfs_gateway_url = 'https://ipfs.infura.io:5001'
                            self.log.info('Using default IPFS Infura gateway URL {ipfs_gateway_url}',
                                          ipfs_gateway_url=hlid(ipfs_gateway_url))

                        ipfs_config_url = '{}/api/v0/cat?arg={}&encoding=json'.format(
                            ipfs_gateway_url, xbr_node_config)
                        resp = requests.get(ipfs_config_url)

                        xbr_node_config_data = resp.json()

                        from pprint import pprint
                        pprint(xbr_node_config_data)

                        self.personality.check_config(self.personality, xbr_node_config_data)

                        if 'controller' not in xbr_node_config_data:
                            xbr_node_config_data['controller'] = {}

                        if 'id' not in xbr_node_config_data['controller']:
                            xbr_node_config_data['controller']['id'] = str(uuid.UUID(bytes=xbr_node_id))
                            self.log.info('Deriving node ID {node_id} from XBR network node ID',
                                          node_id=hlid(xbr_node_config_data['controller']['id']))
                        else:
                            self.log.info('Setting node ID {node_id} from XBR network node configuration',
                                          node_id=hlid(xbr_node_config_data['controller']['id']))

                        self._config = xbr_node_config_data

                        self.log.info('Node configuration loaded from XBR network (xbr_node_id="{xbr_node_id}")',
                                      xbr_node_id=hlid('0x' + binascii.b2a_hex(xbr_node_id).decode()))
                    else:
                        self.log.debug(
                            'There is no node configuration stored in XBR network (xbr_node_id="{xbr_node_id}")',
                            xbr_node_id=hlid('0x' + binascii.b2a_hex(xbr_node_id).decode()))

                    config_source = self.CONFIG_SOURCE_XBRNETWORK
                else:
                    self.log.warn('Could not find node public key on blockchain yet')
            else:
                self.log.warn('Could not connect to Ethereum blockchain')

        # 4/4: check and set the final merged master node configuration
        #
        self.personality.check_config(self.personality, config)
        self._config = config

        self.log.debug('Master node config after (effective after merge):\n{config}', config=pformat(config))

        return config_source, config_path

    @inlineCallbacks
    def start(self, node_id=None):
        self.log.info('{note} [{method}]',
                      note=hl('Starting node (initialize master-node personality) ..', color='green', bold=True),
                      method=hltype(FabricCenterNode.start))
        res = yield node.FabricNode.start(self, node_id)
        return res

    @inlineCallbacks
    def boot(self):
        self.log.info('Booting node {method}', method=hltype(FabricCenterNode.boot))

        # get fabric controller configuration
        #
        controller_config_extra = self._config.get('controller', {}).get('fabric-center', {})

        # apply node config
        #
        yield self.boot_from_config(self._config)
        self._local_config_applied = True

        # start node manager
        #
        extra = {
            'cbdir': self._cbdir,
            'database': controller_config_extra.get('database', {}),
            'ready': Deferred(),
        }
        config = ComponentConfig(self._realm, extra)
        self._bridge_session = FabricServiceNodeManager(config=config,
                                                        reactor=self._reactor,
                                                        personality=self.personality,
                                                        node=self)
        router = self._router_factory.get(self._realm)
        self._router_session_factory.add(self._bridge_session, router, authrole='trusted')
        yield extra['ready']

    def _add_extra_controller_components(self, controller_options):
        self.log.debug('FabricCenterNode._add_extra_controller_components: no extra controller components added')

    def _set_shutdown_triggers(self, controller_options):
        # CFC workers are not supposed to exit with error, so we shutdown when a worker
        # exits with error (and rely on systemd to restart CFC) - OR if we explicitly
        # ask the node to shutdown (eg CTRL-C it)

        # self._node_shutdown_triggers = [checkconfig.NODE_SHUTDOWN_ON_SHUTDOWN_REQUESTED]
        # self._node_shutdown_triggers = [checkconfig.NODE_SHUTDOWN_ON_WORKER_EXIT]
        # self._node_shutdown_triggers = [checkconfig.NODE_SHUTDOWN_ON_WORKER_EXIT_WITH_ERROR]
        self._node_shutdown_triggers = [checkconfig.NODE_SHUTDOWN_ON_LAST_WORKER_EXIT]

    def _add_worker_role(self, worker_auth_role, options):
        worker_role_config = {
            # each (native) worker is authenticated under a worker-specific authrole
            "name":
            worker_auth_role,
            "permissions": [
                # the worker requires these permissions to work:
                {
                    # management API provided by the worker. note that the worker management API is provided under
                    # the URI prefix "crossbar.worker.<worker_id>". note that the worker is also authenticated
                    # under authrole <worker_auth_role> on realm "crossbar"
                    "uri": worker_auth_role,
                    "match": "prefix",
                    "allow": {
                        "call": True,
                        "register": True,
                        "publish": True,
                        "subscribe": True
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                },
                {
                    # controller procedure called by the worker (to check for controller status)
                    "uri": "crossbar.get_status",
                    "match": "exact",
                    "allow": {
                        "call": True,
                        "register": False,
                        "publish": False,
                        "subscribe": False
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                }
            ]
        }
        # if configured to expose the controller connection within the worker (to make it available
        # in user code such as dynamic authenticators and router/container components), also add
        # permissions to actually use the (local) node management API
        if options.get('expose_controller', True):
            vendor_permissions = {
                "uri": "crossbar.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": False,
                    "publish": False,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }
            worker_role_config["permissions"].append(vendor_permissions)
            vendor_permissions = {
                "uri": "crossbarfabriccenter.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": True,
                    "publish": True,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }
            worker_role_config["permissions"].append(vendor_permissions)

        self._router_factory.add_role(self._realm, worker_role_config)

        self.log.info('worker-specific role "{authrole}" added on node management router realm "{realm}" {func}',
                      func=hltype(self._add_worker_role),
                      authrole=hlid(worker_role_config['name']),
                      realm=hlid(self._realm))



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.web.http_headers import Headers
from twisted.internet.defer import inlineCallbacks

from autobahn.wamp.types import ComponentConfig, PublishOptions

from crossbar.test import TestCase
from crossbar.bridge.rest.test import MockTransport, MockWebTransport
from crossbar.bridge.rest import MessageForwarder


class MessageForwarderTestCase(TestCase):
    @inlineCallbacks
    def test_basic_web(self):
        """
        Plain request, no params.
        """
        extra = {"subscriptions": [{"url": "https://foo.com/msg", "topic": "io.crossbar.forward1"}]}
        config = ComponentConfig(realm="realm1", extra=extra)

        m = MockWebTransport(self)
        m._addResponse(200, "whee")

        c = MessageForwarder(config=config, webTransport=m)
        MockTransport(c)

        res = yield c.publish("io.crossbar.forward1", "hi", options=PublishOptions(acknowledge=True))

        self.assertNotEqual(res.id, None)
        self.assertEqual(m.maderequest["args"], ("POST", b"https://foo.com/msg"))
        self.assertEqual(m.maderequest["kwargs"], {
            "data": b'{"args":["hi"],"kwargs":{}}',
            "headers": Headers({b"Content-Type": [b"application/json"]})
        })



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.trial.unittest import TestCase

from crossbar.bridge.mqtt.wamp import _mqtt_topicfilter_to_wamp, _mqtt_topicname_to_wamp


class MQTTTopicTests(TestCase):
    """
    Tests our processing of MQTT topic names and topic filters.
    """
    def test_topic_filters_valid(self):
        """
        Test checking and conversion of valid MQTT topic filters.
        """
        for topic, mapped_topic, match_policy in [('foobar', 'foobar', 'exact'),
                                                  ('com/example/topic1', 'com.example.topic1', 'exact'),
                                                  ('tennis/player1', 'tennis.player1', 'exact'),
                                                  ('tennis', 'tennis', 'exact'),
                                                  ('sport/tennis/player1/#', 'sport.tennis.player1.', 'prefix'),
                                                  ('#', '', 'prefix'), ('+', '', 'wildcard'),
                                                  ('+/tennis/player1', '.tennis.player1', 'wildcard'),
                                                  ('sport/+/player1', 'sport..player1', 'wildcard'),
                                                  ('sport/tennis/+', 'sport.tennis.', 'wildcard'),
                                                  ('+/+/+', '..', 'wildcard')]:
            _mapped_topic, _match_policy = _mqtt_topicfilter_to_wamp(topic)
            self.assertEqual(_mapped_topic, mapped_topic)
            self.assertEqual(_match_policy, match_policy)

    def test_topic_filters_invalid(self):
        """
        Test invalid MQTT topic filters.
        """
        for topic in [
                # invalid types
                None,
                23,
                b'bla',

                # invalid according to MQTT v3.3.1
                '',
                'sport/tennis#',
                'sport/tennis/#/ranking',
                '##',
                '++',
                'sport+',
                'sport/+ab/player1',

                # The following are valid topic names/filters in MQTT v3.3.1,
                # but are invalid in Crossbar.io (WAMP):

                # cannot combine wildcard and prefix matching
                '+/tennis/#',
                '+/#',

                # components cannot contain whitespace
                'spo oh rt',
                'foo/b r/baz',
                'sport/ /#',
                'sport/ten nis/#',

                # no trailing/leading level separator
                '/tennis/player1',
                'tennis/player1/',
                '/tennis/player1/',

                # no double, triple, .. level separator
                'tennis//player1',
                '//',
                'tennis///player1',
                '///',
        ]:
            with self.assertRaises(TypeError):
                _mqtt_topicfilter_to_wamp(topic)

    def test_topic_names_valid(self):
        """
        Test checking and conversion of valid MQTT topic names.
        """
        for topic, mapped_topic in [
            ('com/example/topic1', 'com.example.topic1'),
            ('a', 'a'),
            ('foobar', 'foobar'),
            ('com/example/topic1', 'com.example.topic1'),
            ('tennis/player1', 'tennis.player1'),
            ('tennis', 'tennis'),
        ]:
            _mapped_topic = _mqtt_topicname_to_wamp(topic)
            self.assertEqual(_mapped_topic, mapped_topic)

    def test_topic_names_invalid(self):
        """
        Test invalid MQTT topic names.
        """
        for topic in [
                # invalid types
                None,
                23,
                b'bla',

                # must not contain wildcard characters
                '#',
                '+',
                'sport/tennis/player1/#',
                'sport/+/player1',

                # topic cannot be empty
                '',

                # The following are valid topic names/filters in MQTT v3.3.1,
                # but are invalid in Crossbar.io (WAMP):

                # components cannot contain whitespace
                'spo oh rt',
                'foo/b r/baz',

                # no trailing/leading level separator
                '/tennis/player1',
                'tennis/player1/',
                '/tennis/player1/',

                # no double, triple, .. level separator
                'tennis//player1',
                '//',
                'tennis///player1',
                '///',
        ]:
            with self.assertRaises(TypeError):
                _mqtt_topicname_to_wamp(topic)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import six

from txaio import make_logger

from twisted.internet import reactor

from autobahn import wamp
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import PublishOptions

from crossbar._util import hl, hltype

from crossbar.worker.router import RouterController
from crossbar.worker.types import RouterRealm
from crossbar.router.router import RouterFactory, Router

from crossbar.edge.worker.tracing import FabricRouterTrace

__all__ = ('ExtRouterController', )


class RouterRealmInterface(object):
    def __init__(self, id, config):
        self.id = id
        self.config = config


class RouterInterface(object):
    def __init__(self, router, uri):
        self.router = router
        self.uri = uri


class ExtRouter(Router):
    """
    Router extended with crossbar features.
    """
    def __init__(self, factory, realm, options=None, store=None):
        Router.__init__(self, factory, realm, options, store)
        self._interfaces = {}

    def has_interface(self, uri):
        return uri in self._interfaces

    def add_interface(self, interface):
        overwritten = interface.uri in self._interfaces
        self._interfaces[interface.uri] = interface
        return overwritten

    def drop_interface(self, interface):
        if interface.uri in self._interfaces:
            del self._interfaces[interface.uri]
            return True
        return False

    def validate(self, payload_type, uri, args, kwargs):
        # Do validation here
        super().validate(payload_type, uri, args, kwargs)


class ExtRouterFactory(RouterFactory):
    """
    Router factory extended with crossbar features.
    """
    router = ExtRouter

    def __init__(self, node_id, worker, options=None):
        RouterFactory.__init__(self, node_id, worker, options=options)

    def add_interface(self, realm, interface):
        assert (type(realm) == six.text_type)
        assert (realm in self._routers)

        router = self._routers[realm]

        router.add_interface(RouterInterface(router, interface['uri']))

    def drop_interface(self, realm, interface_id):
        assert (type(realm) == six.text_type)
        assert (type(interface_id) == six.text_type)

        if realm not in self._routers:
            raise Exception('no router started for realm "{}"'.format(realm))

        router = self._routers[realm]

        if interface_id not in router._interfaces:
            raise Exception('no interface "{}" started on router for realm "{}"'.format(interface_id, realm))

        interface_id = router._interfaces[interface_id]
        router.drop_interface(interface_id)


class ExtRouterRealm(RouterRealm):
    """
    Router realm run-time representation, extended with crossbar features:

    1. router links
    2. interfaces
    """
    def __init__(self, controller, realm_id, config, router=None, session=None):
        """

        :param controller:

        :param realm_id: The realm ID within the router.
        :type realm_id: str

        :param config: The realm configuration.
        :type config: dict

        :param router: The router (within the router worker) serving the realm.
        :type router: :class:`crossbar.edge.worker.router.ExtRouter`

        :param session: The realm service session.
        :type session: :class:`crossbar.router.service.RouterServiceAgent`
        """
        RouterRealm.__init__(self, controller, realm_id, config, router=router, session=session)

        # FIXME
        self.interfaces = {}

    def marshal(self):
        marshalled = RouterRealm.marshal(self)

        # FIXME
        marshalled['interfaces'] = self.interfaces

        return marshalled


class ExtRouterController(RouterController):
    """
    Controller session for crossbar router workers.
    """

    log = make_logger()

    def __init__(self, config=None, reactor=None, personality=None):
        RouterController.__init__(self, config=config, reactor=reactor, personality=personality)

        # router factory / realm classes to be used
        self.router_factory_class = ExtRouterFactory
        self.router_realm_class = ExtRouterRealm

        # map: trace ID -> RouterTrace
        self._traces = {}

        # when users don't provide a trace_id, draw from this enumerator
        self._next_trace = 1

    @wamp.register(None)
    def start_router_realm(self, realm_id, realm_config, details=None):
        self.log.info('Starting router realm "{realm_id}" {method}',
                      realm_id=realm_id,
                      method=hltype(ExtRouterController.start_router_realm))

        # activate this to test:
        if False and realm_config['name'] == 'realm1':
            self.log.info(hl('Auto-renaming realm1 to realm001', color='green', bold=True))
            realm_config['name'] = 'realm001'

        return RouterController.start_router_realm(self, realm_id, realm_config, details)

    def _next_trace_id(self):
        while True:
            trace_id = u'trace{}'.format(self._next_trace)
            self._next_trace += 1
            if trace_id not in self._traces:
                return trace_id

    def _maybe_trace_rx_msg(self, session, msg):
        if self._traces:
            for trace in self._traces.values():
                trace.maybe_trace_rx_msg(session, msg)

    def _maybe_trace_tx_msg(self, session, msg):
        if self._traces:
            for trace in self._traces.values():
                trace.maybe_trace_tx_msg(session, msg)

    @wamp.register(None)
    def get_traces(self, include_stopped=False, details=None):
        self.log.debug('get_traces(inclue_stopped="{include_stopped}")', include_stopped=include_stopped)

        return sorted(self._traces.keys())

    @wamp.register(None)
    def get_trace(self, trace_id, details=None):
        self.log.debug('get_trace(trace_id="{trace_id}")', trace_id=trace_id)

        if trace_id not in self._traces:
            raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))

        return self._traces[trace_id].marshal()

    @wamp.register(None)
    def start_trace(self, trace_id=None, trace_options=None, details=None):
        self.log.info('start_trace(trace_id="{trace_id}", trace_options="{trace_options}")',
                      trace_id=trace_id,
                      trace_options=trace_options)

        assert (trace_id is None or type(trace_id) == six.text_type)
        assert (trace_options is None or type(trace_options) == dict)

        trace_id = trace_id or self._next_trace_id()
        trace_options = trace_options or {}

        # tracing level: u'message' (default) or u'action'
        trace_level = trace_options.get(u'trace_level', u'message')
        if trace_level not in [u'message', u'action']:
            emsg = 'invalid tracing options: trace_level must be one of ["message", "action"], but was "{}"'.format(
                trace_level)
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # flag to control tracing of app _payload_
        trace_app_payload = trace_options.get(u'trace_app_payload', False)
        if type(trace_app_payload) != bool:
            emsg = 'invalid tracing options: trace_app_payload must be of type bool, was {}'.format(
                type(trace_app_payload))
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # tracing app payload only makes sense for trace_level == u'message'
        if trace_app_payload and trace_level != u'message':
            emsg = 'invalid tracing options: when trace_app_payload is set, trace_level must be "message", but was "{}"'.format(
                trace_level)
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # parameter to control batching of trace records (in ms)
        batching_period = trace_options.get(u'batching_period', 200)
        if type(batching_period) not in six.integer_types or batching_period < 10 or batching_period > 10000:
            emsg = 'invalid tracing options: batching_period must be an integer [10, 10000], was "{}"'.format(
                type(batching_period))
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # flag to control tracing persistence
        persist = trace_options.get(u'persist', False)
        if type(persist) != bool:
            emsg = 'invalid tracing options: persist must be of type bool, was {}'.format(type(persist))
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # parameter to control trace duration (in secs): if given, automatically stop the trace
        # after the given period of time. if not given, the trace runs until stopped explicitly
        duration = trace_options.get(u'duration', None)
        if duration is not None and (type(duration) not in six.integer_types or duration < 1 or duration > 86400):
            emsg = 'invalid tracing options: duration must be an integer [1, 86400], was "{}"'.format(type(duration))
            self.log.error(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # check user provided trace_id
        if trace_id in self._traces:
            emsg = 'could not start trace: a trace with ID "{}" is already running (or starting)'.format(trace_id)
            self.log.error(emsg)
            raise ApplicationError(u'crossbar.error.already_running', emsg)

        def on_trace_period_finished(trace_id, period, trace_batch):
            if trace_level == u'message':
                trace_data = [trace_record.marshal(self._trace_app_payload) for trace_record in trace_batch]
            elif trace_level == u'action':
                trace_data = [traced_action.marshal() for traced_action in trace_batch]
            else:
                raise Exception('logic error')

            self.publish(u'{}.on_trace_data'.format(self._uri_prefix), trace_id, period, trace_data)

        trace = FabricRouterTrace(self,
                                  trace_id,
                                  on_trace_period_finished=on_trace_period_finished,
                                  trace_level=trace_level,
                                  trace_app_payload=trace_app_payload,
                                  batching_period=batching_period,
                                  persist=persist,
                                  duration=duration)
        trace.start()
        self._traces[trace_id] = trace

        if duration:

            def maybe_stop():
                if trace_id in self._traces and self._traces[trace_id]._status == u'running':
                    self.stop_trace(trace_id, details=details)

            reactor.callLater(float(duration), maybe_stop)

        trace_started = trace.marshal()

        self.publish(u'{}.on_trace_started'.format(self._uri_prefix), trace_id, trace_started)

        return trace_started

    @wamp.register(None)
    def stop_trace(self, trace_id, details=None):
        self.log.info('stop_trace(trace_id="{trace_id}")', trace_id=trace_id)

        if trace_id not in self._traces:
            raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))

        trace = self._traces[trace_id]
        trace.stop()

        del self._traces[trace_id]

        trace_stopped = trace.marshal()

        self.publish(u'{}.on_trace_stopped'.format(self._uri_prefix), trace_id, trace_stopped)

        return trace_stopped

    @wamp.register(None)
    def get_trace_data(self, trace_id, from_seq, to_seq=None, limit=None, details=None):
        self.log.debug(
            'get_trace_data(trace_id="{trace_id}", from_seq="{from_seq})", to_seq="{to_seq}", limit="{limit}")',
            trace_id=trace_id,
            from_seq=from_seq,
            to_seq=to_seq,
            limit=limit)

        if trace_id not in self._traces:
            raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))

        limit = limit or 100
        if limit > 10000:
            raise Exception('limit too large')

        return self._traces[trace_id].get_data(from_seq, to_seq, limit)

    @wamp.register(None)
    def start_router_realm_interface(self, realm_id, interface_id, interface_config, details=None):
        if realm_id not in self.realms:
            raise ApplicationError(u"crossbar.error.no_such_object", "No realm with ID '{}'".format(realm_id))

        if interface_id in self.realms[realm_id].interfaces:
            raise ApplicationError(
                u"crossbar.error.already_exists",
                "An interface with ID '{}' already exists in realm with ID '{}'".format(interface_id, realm_id))

        self.realms[realm_id].interfaces[interface_id] = RouterRealmInterface(interface_id, interface_config)

        realm = self.realms[realm_id].config['name']
        self._router_factory.add_interface(realm, interface_config)

        topic = u'{}.on_router_realm_interface_started'.format(self._uri_prefix)
        event = {u'id': interface_id}
        caller = details.caller if details else None
        self.publish(topic, event, options=PublishOptions(exclude=caller))

    @wamp.register(None)
    def stop_router_realm_interface(self, realm_id, interface_id, details=None):
        if realm_id not in self.realms:
            raise ApplicationError(u"crossbar.error.no_such_object", "No realm with ID '{}'".format(realm_id))

        if interface_id not in self.realms[realm_id].interfaces:
            raise ApplicationError(u"crossbar.error.no_such_object",
                                   "No interface with ID '{}' in realm with ID '{}'".format(interface_id, realm_id))

        del self.realms[realm_id].interfaces[interface_id]

    @wamp.register(None)
    def get_router_realm_interface(self, realm_id, details=None):
        if realm_id not in self.realms:
            raise ApplicationError(u"crossbar.error.no_such_object", "No realm with ID '{}'".format(realm_id))
        return self.realms[realm_id].interfaces.values()



# coding=utf8

##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import binascii
from typing import Optional

import zlmdb

from autobahn.wamp import register
from autobahn.util import generate_serial_number
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import CallDetails

from cfxdb.xbrnetwork.schema import Schema


class Authenticator(ApplicationSession):

    ERROR_INVALID_AUTH_REQUEST = 'xbr.network.error.invalid-auth-request'
    ERROR_INVALID_AUTH_REQUEST_MSG = 'invalid parameter(s) in authentication request: {}'

    def __init__(self, config):
        ApplicationSession.__init__(self, config)

        # ZLMDB database configuration
        #
        self._dbpath = os.path.abspath(config.extra.get('dbpath', './.xbrnetwork'))
        self._db = zlmdb.Database(dbpath=self._dbpath, maxsize=2**30, readonly=False, sync=True)
        self._db.__enter__()
        self._schema = Schema.attach(self._db)

        self._pubkey_by_session = {}
        self._member_by_session = {}
        self._sessions_by_member = {}

        with self._db.begin() as txn:
            cnt_user_keys = self._schema.user_keys.count(txn)

        self.log.info('Database opened from {dbpath} (cnt_user_keys={cnt_user_keys})',
                      dbpath=self._dbpath,
                      cnt_user_keys=cnt_user_keys)

    async def onJoin(self, details):
        regs = await self.register(self)
        for reg in regs:
            self.log.info('{klass} registered procedure {proc}', klass=self.__class__.__name__, proc=reg.procedure)

    @register('xbr.network.authenticator.sessions_by_member')
    def _sessions_by_member(self, member_oid: bytes, details: Optional[CallDetails] = None) -> int:
        return self._sessions_by_member.get(member_oid, None)

    @register('xbr.network.authenticator.member_by_session')
    def _pubkey_by_session(self, session_id: int, details: Optional[CallDetails] = None) -> bytes:
        return self._member_by_session.get(session_id, None)

    @register('xbr.network.authenticator.pubkey_by_session')
    def _member_by_session(self, session_id: int, details: Optional[CallDetails] = None) -> bytes:
        return self._pubkey_by_session.get(session_id, None)

    @register('xbr.network.authenticator.authenticate')
    async def _authenticate(self, realm, authid, details):
        self.log.info('{klass}.authenticate(realm="{realm}", authid="{authid}", details={details})',
                      klass=self.__class__.__name__,
                      realm=realm,
                      authid=authid,
                      details=details)

        if 'authmethod' not in details:
            msg = 'missing "authmethod" in authentication details (WAMP HELLO message details)'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        authmethod = details['authmethod']

        if authmethod not in ['cryptosign']:
            msg = 'authmethod "{}" not permissible'.format(authmethod)
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        if 'authextra' not in details or 'pubkey' not in details['authextra']:
            msg = 'missing public key in authextra for authmethod cryptosign'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))
        pubkey = details['authextra']['pubkey']
        pubkey_raw = binascii.a2b_hex(pubkey)
        assert type(pubkey_raw) == bytes and len(pubkey_raw) == 32

        session_id = details['session']
        assert type(session_id) == int

        with self._db.begin() as txn:
            # double check (again) for username collision, as the mailgun email submit happens async in above after
            # we initially checked for collision
            user_key = self._schema.user_keys[txn, pubkey_raw]
            if user_key:
                account = self._schema.accounts[txn, user_key.owner]
                authrole = 'member'
                # authid = account.username
                authid = 'member-{}'.format(account.oid)
            else:
                account = None
                authrole = 'anonymous'
                authid = 'anonymous-{}'.format(generate_serial_number())

            self._pubkey_by_session[session_id] = pubkey_raw
            if account:
                self._member_by_session[session_id] = account.oid
                if account.oid not in self._sessions_by_member:
                    self._sessions_by_member[account.oid] = []
                self._sessions_by_member[account.oid].append(session_id)

        auth = {'pubkey': pubkey, 'realm': realm, 'authid': authid, 'role': authrole, 'extra': None, 'cache': True}

        self.log.info('{klass}.authenticate(..) => {auth}', klass=self.__class__.__name__, auth=auth)

        return auth



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import json
from typing import Union, Dict, Any

import txaio
from txaio import make_logger

from autobahn import util
from autobahn.wamp import auth
from autobahn.util import hltype, hlval
from autobahn.wamp.types import Accept, Deny, HelloDetails, Challenge, TransportDetails

from crossbar.router.auth.pending import PendingAuth
from crossbar.interfaces import IRealmContainer, IPendingAuth

__all__ = ('PendingAuthWampCra', )


class PendingAuthWampCra(PendingAuth):
    """
    Pending WAMP-CRA authentication.
    """

    AUTHMETHOD = 'wampcra'

    log = make_logger()

    def __init__(self, pending_session_id: int, transport_details: TransportDetails, realm_container: IRealmContainer,
                 config: Dict[str, Any]):
        super(PendingAuthWampCra, self).__init__(
            pending_session_id,
            transport_details,
            realm_container,
            config,
        )

        # filled with what the client requested originally (if any)
        self._realm = None
        self._authid = None

        # The signature we expect the client to send in AUTHENTICATE.
        self._signature = None

    def _compute_challenge(self, user):
        """
        Returns: challenge, signature
        """
        challenge_obj = {
            'authid': self._authid,
            'authrole': self._authrole,
            'authmethod': self._authmethod,
            'authprovider': self._authprovider,
            'session': self._session_details['session'],
            'nonce': util.newid(64),
            'timestamp': util.utcnow()
        }
        challenge: str = json.dumps(challenge_obj, ensure_ascii=False)
        secret = user['secret'].encode('utf8')
        signature = auth.compute_wcs(secret, challenge.encode('utf8')).decode('ascii')

        # extra data to send to client in CHALLENGE
        extra = {'challenge': challenge}

        # when using salted passwords, provide the client with
        # the salt and then PBKDF2 parameters used
        if 'salt' in user:
            extra['salt'] = user['salt']
            extra['iterations'] = user.get('iterations', 1000)
            extra['keylen'] = user.get('keylen', 32)

        return extra, signature

    def hello(self, realm: str, details: HelloDetails) -> Union[Accept, Deny, Challenge]:

        # remember the realm the client requested to join (if any)
        self._realm = realm

        # remember the authid the client wants to identify as (if any)
        self._authid = details.authid

        # define local helpers
        if self._config['type'] in ['dynamic', 'function']:

            def on_authenticate_ok(_principal):
                _error = self._assign_principal(_principal)
                if _error:
                    return _error

                # now compute CHALLENGE.Extra and signature expected
                _extra, self._signature = self._compute_challenge(_principal)
                return Challenge(self._authmethod, _extra)

            def on_authenticate_error(err):
                return self._marshal_dynamic_authenticator_error(err)

        # use static principal database from configuration
        if self._config['type'] == 'static':

            self._authprovider = 'static'

            if self._authid in self._config.get('users', {}):

                principal = self._config['users'][self._authid]

                error = self._assign_principal(principal)
                if error:
                    return error

                # now compute CHALLENGE.Extra and signature as
                # expected for WAMP-CRA
                extra, self._signature = self._compute_challenge(principal)

                return Challenge(self._authmethod, extra)
            else:
                return Deny(message='no principal with authid "{}" exists'.format(details.authid))

        # use configured procedure to dynamically get a ticket for the principal
        elif self._config['type'] == 'dynamic':

            self._authprovider = 'dynamic'

            init_d = txaio.as_future(self._init_dynamic_authenticator)

            def init(result):
                if result:
                    return result

                self._session_details['authmethod'] = self._authmethod  # from AUTHMETHOD, via base
                self._session_details['authid'] = details.authid
                self._session_details['authrole'] = details.authrole
                self._session_details['authextra'] = details.authextra

                d = self._authenticator_session.call(self._authenticator, realm, details.authid, self._session_details)
                d.addCallbacks(on_authenticate_ok, on_authenticate_error)

                return d

            init_d.addBoth(init)
            return init_d

        elif self._config['type'] == 'function':

            self._authprovider = 'function'

            init_d = txaio.as_future(self._init_function_authenticator)

            def init(result):
                if result:
                    return result

                self._session_details['authmethod'] = self._authmethod  # from AUTHMETHOD, via base
                self._session_details['authid'] = details.authid
                self._session_details['authrole'] = details.authrole
                self._session_details['authextra'] = details.authextra

                auth_d = txaio.as_future(self._authenticator, realm, details.authid, self._session_details)
                auth_d.addCallbacks(on_authenticate_ok, on_authenticate_error)

                return auth_d

            init_d.addBoth(init)
            return init_d

        else:
            # should not arrive here, as config errors should be caught earlier
            return Deny(message='invalid authentication configuration (authentication type "{}" is unknown)'.format(
                self._config['type']))

    def authenticate(self, signature: str) -> Union[Accept, Deny]:

        if signature == self._signature:
            # signature was valid: accept the client
            return self._accept()
        else:
            # signature was invalid: deny the client
            self.log.warn('{func}: WAMP-CRA client signature is invalid (expected {expected} but got {signature})',
                          func=hltype(self.authenticate),
                          expected=hlval(self._signature),
                          signature=hlval(signature, color='red'))
            return Deny(message='WAMP-CRA client signature is invalid')


IPendingAuth.register(PendingAuthWampCra)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import psutil

from crossbar.edge.worker.monitor._base import Monitor
from txaio import perf_counter_ns

__all__ = ('StorageMonitor', )


class StorageMonitor(Monitor):
    """
    Storage and disk IO monitoring via psutils.
    """

    ID = u'storage'

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        hdata = Monitor.poll(self)

        start = perf_counter_ns()

        devices = {}
        usage = {}
        counters = psutil.disk_io_counters(True)
        for dev in psutil.disk_partitions():
            if dev.device.startswith('/dev/loop'):
                continue
            key = dev.device.split('/')[-1]
            if key not in devices:
                if key in counters:
                    devices[key] = dict(dev._asdict(), **counters[key]._asdict())
                else:
                    devices[key] = dev._asdict()
                usage[key] = psutil.disk_usage(dev.mountpoint)._asdict()

        hdata['devices'] = devices
        hdata['usage'] = usage

        hdata[u'elapsed'] = perf_counter_ns() - start

        self._last_value = hdata

        return hdata



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from crossbar.edge.worker.xbr._marketmaker import MarketMaker
from crossbar.edge.worker.xbr._authenticator import Authenticator

__all__ = ('MarketMaker', 'Authenticator')



import os
import sys
import venv
import json
from os.path import abspath
from os.path import join
from subprocess import check_call  # nosec

import click

from crossbar.quickstart.main import hl  # XXX "util" or similar instead?

__all__ = ('main', )


@click.command()
@click.option("--framework",
              prompt=True,
              type=click.Choice(['twisted', 'asyncio']),
              help="The Python networking framework to use",
              default="twisted")
@click.option(
    "--directory",
    prompt=True,
    default="./cbquickstart",
    help="Directory to create and put examples into",
    type=click.Path(exists=False),
)
@click.option(
    "--cfc",
    default=None,
    help="Master node connection",
    metavar="URI",
)
def main(framework, directory, cfc):
    """
    """
    directory = abspath(directory)
    if cfc is None:
        cfc = "ws://localhost:4444"

    click.echo('{cb} Project Quickstart (virtualenv)\n'.format(cb=hl('Crossbar.io / XBR'), ))
    click.echo("  Framework:  {fw}\n"
               "  Location:   {d}\n"
               "  CFC URI:    {fabric_uri}\n".format(
                   fw=framework,
                   d=directory,
                   fabric_uri=cfc,
               ))

    os.mkdir(directory)
    venv_location = join(directory, 'venv')
    _create_virtualenv(venv_location, framework)
    _create_applications(directory, venv_location, cfc)
    _create_makefile(directory, venv_location)

    click.echo("\n"
               "Quickstart Successful\n"
               "\n"
               "There is Makefile in '{mf}'\n"
               "Starting one shell per target, run these (in this order):\n"
               "  - make etcd\n"
               "  - make cfc\n"
               "  - make app0\n".format(mf=click.style(directory, bold=True, fg='yellow'), ))


def _create_makefile(directory, venv_location):
    template_args = {
        "top_level": directory,
        "venv": venv_location,
        "command_line": " ".join(sys.argv),
    }
    os.mkdir(join(directory, "etcd-data"))
    with open(join(directory, "Makefile"), "w") as makefile:
        makefile.write("""
# generated makefile used command:
# {command_line}

.PHONY: run cfc app0 etcd

run: etcd cfc app0
\techo "running"

etcd:
\tdocker run \
\t\t--rm \
\t\t-p 2379:2379 \
\t\t-p 2380:2380 \
\t\t-v /usr/share/ca-certificates/:/etc/ssl/certs \
\t\t-v {top_level}/etcd-data:/etcd-data \
\t\t--name etcd \
\t\t--net=host \
\t\tquay.io/coreos/etcd:latest \
\t\t\t/usr/local/bin/etcd \
\t\t\t--data-dir=/etcd-data \
\t\t\t--name cf-etcd \
\t\t\t--advertise-client-urls http://0.0.0.0:2379 \
\t\t\t--listen-client-urls http://0.0.0.0:2379

cfc:
\t{venv}/bin/crossbar master start --cbdir {top_level}/master/.crossbar

app0:
\t{venv}/bin/crossbar edge start --cbdir {top_level}/app0/.crossbar
""".format(**template_args))


def _create_applications(location, venv, cfc):
    cbfx = join(venv, "bin", "crossbar")
    app0 = join(location, "app0")
    master = join(location, "master")

    # create a CFC
    check_call([cbfx, "master", "init", "--appdir", master])  # nosec
    with open(join(master, ".crossbar", "config.json")) as f:
        data = json.load(f)
    data['workers'][0]['transports'][0]['endpoint']['port'] = 4444
    with open(join(master, ".crossbar", "config.json"), 'w') as f:
        json.dump(data, f, indent=4)

    # create an 'edge' node (application)
    check_call([cbfx, "edge", "init", "--appdir", app0])  # nosec
    # edit configuration
    with open(join(app0, ".crossbar", "config.json")) as f:
        data = json.load(f)

    if cfc:
        data['controller']['fabric'] = {
            'transport': {
                'url': cfc,
            }
        }
    with open(join(app0, ".crossbar", "config.json"), 'w') as f:
        json.dump(data, f, indent=4)


def _create_virtualenv(location, framework):
    """
    Create a new virtual environment.

    :param location: a non-existant directory (parent must exist though)
    """
    click.echo(click.style("\ncreating '{env}'\n".format(env=location), bold=True, fg="yellow"))
    venv.create(location, with_pip=True)

    pip = join(location, "bin", "pip")

    click.echo(click.style("\nupgrading pip\n", bold=True, fg="yellow"))
    check_call([pip, "install", "--upgrade", "pip"])  # nosec

    click.echo(click.style("\ninstalling software\n", bold=True, fg="yellow"))

    check_call([pip, "install", "autobahn[{}]".format(framework)])  # nosec
    check_call([pip, "install", "crossbar"])  # nosec

    # XXX I guess we just assume we're "in" a Fabric checkout, because
    # how else would you even be running this command?
    check_call([pip, "install", "--editable", "."])  # nosec
    # XXX FIXME why isn't ^ installing our dependencies??
    check_call([pip, "install", "-r", "requirements-min.txt"])  # nosec



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from autobahn.twisted.wamp import WampRawSocketClientFactory, ApplicationSessionFactory
# from autobahn.twisted.wamp import WampWebSocketClientFactory
from autobahn.wamp.types import ComponentConfig

from crossbar.router.router import RouterFactory
from crossbar.router.session import RouterSessionFactory
from crossbar.router.service import RouterServiceAgent
from crossbar.worker.types import RouterRealm
from crossbar.router.role import RouterRoleStaticAuth
from crossbar.router.protocol import WampRawSocketServerFactory, WampWebSocketServerFactory
from crossbar.router.unisocket import UniSocketServerFactory

from twisted.test.iosim import connect, FakeTransport


def make_router():
    """
    Make a router, and return it and a RawSocket factory.
    """
    # create a router factory
    router_factory = RouterFactory('node1', 'worker1', None)

    # create a router session factory
    session_factory = RouterSessionFactory(router_factory)

    # Create a new WebSocket factory
    websocket_server_factory = WampWebSocketServerFactory(session_factory, '.', {}, None)

    # Create a new RawSocket factory
    rawsocket_server_factory = WampRawSocketServerFactory(session_factory, {})

    # Create a new UniSocket factory
    server_factory = UniSocketServerFactory(websocket_factory_map={'/': websocket_server_factory},
                                            rawsocket_factory=rawsocket_server_factory)

    return router_factory, server_factory, session_factory


def add_realm_to_router(router_factory, session_factory, realm_name='default', realm_options={}):

    opts = dict(realm_options)
    opts.update({'name': realm_name})

    # start a realm
    realm = RouterRealm(None, None, opts)
    router = router_factory.start_realm(realm)

    extra = {}
    session_config = ComponentConfig(realm_name, extra)
    realm.session = RouterServiceAgent(session_config, router)

    # allow everything
    default_permissions = {
        'uri': '',
        'match': 'prefix',
        'allow': {
            'call': True,
            'register': True,
            'publish': True,
            'subscribe': True
        }
    }

    router = router_factory.get(realm_name)
    router.add_role(RouterRoleStaticAuth(router, 'anonymous', default_permissions=default_permissions))

    session_factory.add(realm.session, router, authrole='trusted')

    return router


def make_router_and_realm(realm_name='default'):

    router_factory, server_factory, session_factory = make_router()
    router = add_realm_to_router(router_factory, session_factory, realm_name)
    return router, server_factory, session_factory


def connect_application_session(server_factory, MyApplicationSession, component_config=None):
    """
    Connect an ApplicationSession class to the given server factory.
    """
    application_session_factory = ApplicationSessionFactory(component_config)
    application_session_factory.session = MyApplicationSession

    client_factory = WampRawSocketClientFactory(application_session_factory)
    # client_factory = WampWebSocketClientFactory(application_session_factory)

    server_protocol = server_factory.buildProtocol(None)
    client_protocol = client_factory.buildProtocol(None)

    server_transport = FakeTransport(server_protocol, True)
    client_transport = FakeTransport(client_protocol, False)

    pump = connect(server_protocol, server_transport, client_protocol, client_transport, debug=False)

    return client_protocol._session, pump



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import uuid

from autobahn import util

from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

import humanize
from datetime import datetime

import zlmdb

from crossbar._util import hl, hlid, hltype
from cfxdb.globalschema import GlobalSchema
from cfxdb.user import User, UserRole
from cfxdb.user import ActivationType, ActivationStatus, ActivationToken
from crossbar.master.node.messenger import Messenger

__all__ = ('Authenticator', 'Principal')


class Principal(object):
    """
    """
    def __init__(self, realm, authid, role, extra):
        """
        """
        self.realm = realm
        self.authid = authid
        self.role = role
        self.extra = extra

    @staticmethod
    def parse(obj):
        realm = obj.get('realm', None)
        authid = obj.get('authid', None)
        role = obj.get('role', None)
        extra = obj.get('extra', None)
        return Principal(realm, authid, role, extra)


class Node(object):
    def __init__(self, node_id, heartbeat, heartbeat_time, status='online'):
        self.node_id = node_id
        self.heartbeat = heartbeat
        self.heartbeat_time = heartbeat_time
        self.status = status

    def marshal(self):
        return {
            'id': self.node_id,
            'heartbeat': self.heartbeat,
            'timestamp': self.heartbeat_time,
            'status': self.status,
        }

    @staticmethod
    def parse(data):
        assert (type(data) == dict)

        node_id = data.get('node_id', None)
        heartbeat = data.get('heartbeat', None)
        heartbeat_time = data.get('heartbeat_time', None)
        status = data.get('status', None)

        # FIXME: check above

        node = Node(node_id, heartbeat, heartbeat_time, status)
        return node


class Authenticator(ApplicationSession):
    """
    Central CFC dynamic authenticator. This component is responsible for all
    frontend WAMP connections to CFC, for both CF nodes, CFC UI and user CFC scripts.
    """

    GLOBAL_USER_REALM = 'com.crossbario.fabric'
    """
    Global users realm on Crossbar.io.
    """

    GLOBAL_USER_REALM_USER_ROLE = 'user'
    """
    The WAMP authrole regular users get on the Crossbar.io domain (global) users
    realm. A role different from this only makes sense for Crossbar.io admins.
    """

    MREALM_USER_ROLES = ['guest', 'developer', 'operator', 'admin', 'owner']
    """
    All permissible roles a user can take on a management realm. This is a fixed
    set hardwired into Crossbar.io!
    """

    MREALM_CREATOR_DEFAULT_ROLES = ['owner']
    """
    THe set of roles a user creating a new management realm gets by default.
    """

    MREALM_NODE_ROLE = 'node'
    """
    The (fixed) role a user node gets when joining the management realm it is
    paired to.
    """

    ERROR_AUTH_INVALID_PARAMETERS = 'fabric.auth-failed.invalid-parameters'
    ERROR_AUTH_INVALID_PARAMETERS_MSG = 'Invalid parameters in authentication: {}'

    ERROR_AUTH_PENDING_ACT = 'fabric.auth-failed.pending-activation'
    ERROR_AUTH_PENDING_ACT_MSG = 'There is a pending activation (from {} ago) - please check your email inbox, or request a new code'

    ERROR_AUTH_NO_PENDING_ACT = 'fabric.auth-failed.no-pending-activation'
    ERROR_AUTH_NO_PENDING_ACT_MSG = 'There is no (pending) activation for this user/pubkey, but an activation code was provided'

    ERROR_AUTH_INVALID_ACT_CODE = 'fabric.auth-failed.invalid-activation-code'
    ERROR_AUTH_INVALID_ACT_CODE_MSG = 'This activation code is invalid: {}'

    ERROR_AUTH_NODE_UNPAIRED = 'fabric.auth-failed.node-unpaired'
    ERROR_AUTH_NODE_UNPAIRED_MSG = 'This node is unpaired. Please pair the node with management realm first.'

    ERROR_AUTH_NODE_ALREADY_CONNECTED = 'fabric.auth-failed.node-already-connected'
    ERROR_AUTH_NODE_ALREADY_CONNECTED_MSG = 'A node with this pubkey/node_id/authid is already connected.'

    ERROR_AUTH_EMAIL_FAILURE = 'fabric.auth-failed.email-failure'

    ERROR_AUTH_NEW_USER = 'fabric.auth-failed.new-user-auth-code-sent'
    ERROR_AUTH_NEW_USER_MSG = 'We have sent an authentication code to {email}.'

    ERROR_AUTH_REGISTERED_USER = 'fabric.auth-failed.registered-user-auth-code-sent'
    ERROR_AUTH_REGISTERED_USER_MSG = 'We have sent an authentication code to {email}.'

    def __init__(self, config):
        ApplicationSession.__init__(self, config)
        self._connected_nodes = {}

        # public keys of superusers
        self._superusers = []
        if 'CROSSBAR_FABRIC_SUPERUSER' in os.environ:
            superuser_keyfile = os.environ['CROSSBAR_FABRIC_SUPERUSER'].strip()
            if superuser_keyfile:
                pubkey_file = os.path.abspath(os.environ['CROSSBAR_FABRIC_SUPERUSER'])
                if not os.path.exists(pubkey_file):
                    raise Exception(
                        'superuser public key file {} (set from CROSSBAR_FABRIC_SUPERUSER env var) does not exist'.
                        format(pubkey_file))
                with open(pubkey_file, 'r') as f:
                    data = f.read()
                    pubkey_hex = None
                    for line in data.splitlines():
                        if line.startswith('public-key-ed25519'):
                            pubkey_hex = line.split(':')[1].strip()
                            break
                    if pubkey_hex is None:
                        raise Exception(
                            'no public key line found in super user public key file {}'.format(pubkey_file))

                    self._superusers.append(pubkey_hex)
                    self.log.info(
                        hl('SUPERUSER public key {} loaded from {}'.format(pubkey_hex, pubkey_file),
                           color='green',
                           bold=True))

    async def onJoin(self, details):
        """
        Authenticator has joined the global realm ("com.crossbario.fabric").

        :param details: Session details.
        :type details: :class:`autobahn.wamp.types.SessionDetails`
        """
        # create database and attach tables to database slots
        #
        cbdir = self.config.extra['cbdir']
        config = self.config.extra.get('database', {})

        dbpath = config.get('path', '.db-controller')
        assert type(dbpath) == str
        dbpath = os.path.join(cbdir, dbpath)

        maxsize = config.get('maxsize', 128 * 2**20)
        assert type(maxsize) == int
        # allow maxsize 128kiB to 128GiB
        assert maxsize >= 128 * 1024 and maxsize <= 128 * 2**30

        self.log.debug(
            '{msg} [realm={realm}, session={session}, authid={authid}, authrole={authrole}, dbpath={dbpath}, maxsize={maxsize}]',
            realm=hlid(details.realm),
            session=details.session,
            authid=details.authid,
            authrole=hlid(details.authrole),
            msg=hl('Crossbar.io main authenticator starting ..', bold=True, color='green'),
            dbpath=hlid(dbpath),
            maxsize=hlid(maxsize))

        self._db = zlmdb.Database(dbpath=dbpath, maxsize=maxsize, readonly=False, sync=True)
        self._db.__enter__()
        self._schema = GlobalSchema.attach(self._db)

        # Mailgun access key
        #
        access_key = None
        if 'mailgun' in self.config.extra and \
           'access_key' in self.config.extra['mailgun'] and \
           self.config.extra['mailgun']['access_key']:
            access_key = self.config.extra['mailgun']['access_key']
        else:
            if 'MAILGUN_KEY' in os.environ:
                access_key = os.environ['MAILGUN_KEY']
            else:
                self.log.warn('Mailgun access key unconfigured (not in config, and no env var MAILGUN_KEY set)')

        # Mailgun mail submit URL
        # eg https://api.mailgun.net/v3/mailing.crossbar.io/messages
        #
        submit_url = None
        if 'mailgun' in self.config.extra and \
           'submit_url' in self.config.extra['mailgun'] and \
           self.config.extra['mailgun']['submit_url']:
            submit_url = self.config.extra['mailgun']['submit_url']
        else:
            if 'MAILGUN_URL' in os.environ:
                submit_url = os.environ['MAILGUN_URL']
            else:
                self.log.warn('Mailgun submit URL unconfigured (not in config, and no env var MAILGUN_URL set)')

        # Quick hack - the above code doesn't work properly
        access_key = os.environ.get('MAILGUN_KEY', access_key)
        submit_url = os.environ.get('MAILGUN_URL', submit_url)

        # create Email sender. if submit_url or access_key was None, then the Messenger
        # will not send out mails, but log them at level "warn"!
        #
        self._messenger = Messenger(submit_url, access_key)

        # register our dynamic authenticator
        try:
            await self.register(self._authenticate, 'com.crossbario.fabric.authenticate')
        except Exception:
            self.log.failure('failed to register dynamic authenticator: {log_failure}')
            raise Exception('fatal: failed to register dynamic authenticator')

        # set service session to our authenticator
        self.config.controller.set_service_session(self, self._realm, self._authrole)

        self.log.info(
            'Ok, {component} ready on realm "{realm}" (session={session}, authid="{authid}", authrole="{authrole}", controller={controller}) {func}',
            component=hl('master node authenticator', color='green', bold=True),
            func=hltype(self.onJoin),
            realm=hlid(self._realm),
            session=hlid(self._session_id),
            authid=hlid(self._authid),
            authrole=hlid(self._authrole),
            controller=self.config.controller)

    async def _authenticate(self, realm, authid, details):
        """
        Main authenticator for Crossbar.io. This authenticates both users and user nodes
        to Crossbar.io by authenticating against information stored in controller database.
        """
        self.log.debug('authenticate({realm}, {authid}, {details})', realm=realm, authid=authid, details=details)

        # proceed according to authmethod
        #
        if 'authmethod' not in details:
            msg = 'missing "authmethod" in authentication details (WAMP HELLO message details)'
            raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                   Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))

        authmethod = details['authmethod']

        if authmethod not in ['cryptosign']:
            msg = 'authmethod "{}" not permissible'.format(authmethod)
            raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                   Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))

        if authmethod == 'cryptosign':

            # extract mandatory public key
            #
            if 'authextra' not in details or 'pubkey' not in details['authextra']:
                msg = 'missing public key in authextra for authmethod cryptosign'
                raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                       Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))
            pubkey = details['authextra']['pubkey']

            # check requested authrole
            #
            authrole = details.get('authrole', None)
            if realm is None:
                # this is either a node joining a management realm or a user joining
                # the global user realm. a node must request authrole "node"
                if authrole not in [None, Authenticator.MREALM_NODE_ROLE]:
                    msg = 'invalid requested authrole "{}" for realm "{}"'.format(authrole, realm)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                           Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))

            elif realm == Authenticator.GLOBAL_USER_REALM:
                # this is a user joining the global user realm - authrole is assigned
                # automatically and cannot be requested explicitly
                if authrole not in [None, Authenticator.GLOBAL_USER_REALM_USER_ROLE]:
                    msg = 'invalid requested authrole "{}" for realm "{}"'.format(authrole, realm)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                           Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))
            else:
                # this is a user joining a specific management realm.
                if authrole is not None and authrole not in Authenticator.MREALM_USER_ROLES:
                    msg = 'invalid requested authrole "{}" for realm "{}"'.format(authrole, realm)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                           Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))

            # proceed depending on whether to authenticate a user or a node
            #
            if authrole == Authenticator.MREALM_NODE_ROLE:
                if realm is not None:
                    msg = 'invalid requested realm "{}" for node - nodes MUST NOT request a realm (the realm is auto-assigned based on Crossbar.io configuration)'.format(
                        realm)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                           Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG.format(msg))

                # a node is authenticated based on the pubkey only, and realm/authid/authrole/..
                # are set from the configuration stored in controller database purely
                auth = self._auth_node(pubkey)
                return auth

            else:
                # a user may request to join different realms using the same pubkey, hence
                # we forward all info the client supplied
                activation_code = details['authextra'].get('activation_code', None)
                request_new_activation_code = details['authextra'].get('request_new_activation_code', False)
                auth = await self._auth_user(realm, authid, authrole, pubkey, activation_code,
                                             request_new_activation_code)
                return auth

        # safeguard: should not arrive here
        raise Exception('logic error')

    def _auth_node(self, pubkey):
        """
        Authenticate a Crossbar.io node.

        A user node is purely authenticated based on the node's public key, MUST request
        the authrole="node", and MUST NOT request a specific realm (the management realm
        of a node is automatically assigned by Crossbar.io).
        """
        self.log.debug('authenticating node with pubkey={pubkey}', pubkey=pubkey)

        # lookup the node by its public key (in hex)
        node = None
        with self._db.begin() as txn:
            node_oid = self._schema.idx_nodes_by_pubkey[txn, pubkey]
            if node_oid:
                node = self._schema.nodes[txn, node_oid]

        if node and node.mrealm_oid:
            if node.mrealm_oid not in self._connected_nodes:
                self._connected_nodes[node.mrealm_oid] = {}

            # for nodes, realm + authid MUST be configured, and authextra MAY be configured
            # note that a authenticating node MUST NOT override the configured values for
            # realm, authid, etc  - this is critical. Eg the management realm a node is
            # assigned to must only be chanced under the control of the Crossbar.io
            # backend, since the owner of a management realm pays for the usage the nodes
            # assigned to do produce. And since a node MUST have the realm set (or the whole
            # record is deleted, and the public key "no longer known"), nodes are always
            # authenticated and associated with the correct realm.
            #
            with self._db.begin() as txn:
                mrealm = self._schema.mrealms[txn, node.mrealm_oid]
            if mrealm:

                # use the node name (which can change!)
                assigned_auth_id = node.authid

                # use the node UUID (which never changes)
                # assigned_auth_id = str(node_oid)

                auth = {
                    'pubkey': pubkey,
                    'realm': mrealm.name,
                    'authid': assigned_auth_id,
                    'role': Authenticator.MREALM_NODE_ROLE,
                    'extra': node.authextra,
                    'cache': False
                }

                self._connected_nodes[node.mrealm_oid][node_oid] = auth

                self.log.info(
                    'authenticated managed node "{authid}" with pubkey "{pubkey}" on management realm "{realm}" (authrole="{authrole}") {func}',
                    pubkey=hlid('0x' + pubkey[:16] + '..'),
                    authid=hlid(auth['authid']),
                    authrole=hlid(auth['role']),
                    realm=hlid(auth['realm']),
                    func=hltype(self._auth_node))
                return auth

        # the node pairing process is completely driven from a logged in user session!
        # so either the node already is paired, or it is not - in which case, it needs
        # to be paired by a user, and there is nothing a node could do about or needs
        # to care, so we plainly bail out here ..
        #
        self.log.info('denied unpaired CF node with pubkey {pubkey}', pubkey=pubkey)
        raise ApplicationError(Authenticator.ERROR_AUTH_NODE_UNPAIRED, Authenticator.ERROR_AUTH_NODE_UNPAIRED_MSG)

    async def _auth_user(self,
                         realm,
                         authid,
                         authrole,
                         pubkey,
                         activation_code=None,
                         request_new_activation_code=False):
        """
        Authenticate a Crossbar.io user.

        For a user that is not yet registered, or a user key not yet associated
        with a user, this will raise an ApplicationError signaling the state
        of the registration process.
        """
        self.log.debug(
            'authenticating user for realm={realm}, authid={authid}, authrole={authrole}, pubkey={pubkey}, activation_code={activation_code}, request_new_activation_code={request_new_activation_code}',
            realm=realm,
            authid=authid,
            authrole=authrole,
            pubkey=pubkey,
            activation_code=activation_code,
            request_new_activation_code=request_new_activation_code)

        # we must protect against this!
        if activation_code and not authid:
            raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_PARAMETERS,
                                   Authenticator.ERROR_AUTH_INVALID_PARAMETERS_MSG)

        # get activation for authid/pubkey pair (this will succeed when the user is registered and
        # the pubkey is associated with the user - the 99% case)
        activation = None
        with self._db.begin() as txn:
            oid = self._schema.idx_act_tokens_by_authid_pubkey[txn, authid + pubkey]
            if oid:
                activation = self._schema.activation_tokens[txn, oid]

        # if the activation is still pending, allow to reset it and we will send
        # a new activation code
        if activation and activation.status == ActivationStatus.PENDING and request_new_activation_code:
            with self._db.begin(write=True) as txn:
                oid = self._schema.idx_act_tokens_by_authid_pubkey[txn, authid + pubkey]
                if oid:
                    del self._schema.activation_tokens[txn, oid]
                else:
                    raise Exception('no such activation')
            activation = None

        # get user (if any)
        user = None
        if authid:
            with self._db.begin() as txn:
                oid = self._schema.idx_users_by_email[txn, authid]
                if oid:
                    user = self._schema.users[txn, oid]

        is_new_user = user is None

        self.log.debug('authenticating user={user}, is_new_user={is_new_user}, activation={activation}',
                       user=user,
                       is_new_user=is_new_user,
                       activation=activation)

        if realm is None or realm == 'com.crossbario.fabric':
            realm = Authenticator.GLOBAL_USER_REALM
            authrole = Authenticator.GLOBAL_USER_REALM_USER_ROLE
        else:
            if authrole is None:
                authrole = 'owner'

        # superusers are treated special ..
        if pubkey in self._superusers:
            auth = {
                'pubkey': pubkey,
                'realm': realm,
                'authid': 'superuser',
                'role': authrole,
                'extra': None,
                'cache': False
            }
            self.log.info(
                hl('SUPERUSER authenticated (realm={}, authid={}, authrole={})'.format(
                    auth['realm'], auth['authid'], auth['role']),
                   color='green',
                   bold=True))
            return auth

        # if there is no activation yet, create/store a new one
        elif not activation:

            # check for user provided an activation code, though there is no activation currently
            if activation_code:
                raise ApplicationError(Authenticator.ERROR_AUTH_NO_PENDING_ACT,
                                       Authenticator.ERROR_AUTH_NO_PENDING_ACT_MSG)

            # ok, create a new activation in the database
            activation = ActivationToken()
            activation.oid = uuid.uuid4()
            activation.atype = ActivationType.REGISTRATION if is_new_user else ActivationType.LOGIN
            activation.created = datetime.utcnow()
            # activation.activated = None
            activation.code = util.generate_activation_code()
            activation.status = ActivationStatus.PENDING
            activation.email = authid
            activation.pubkey = pubkey

            with self._db.begin(write=True) as txn:
                self._schema.activation_tokens[txn, activation.oid] = activation

            if is_new_user:
                # send user message with activation code
                await self._messenger.send_user_registration_mail(authid, activation.code)
                self.log.info('User registration mail sent to {authid} with activation code {activation_code}',
                              authid=hlid(authid),
                              activation_code=hl(activation.code, color='red', bold=True))

                # deny authentication by raising an error and providing feedback to client
                raise ApplicationError(Authenticator.ERROR_AUTH_NEW_USER,
                                       Authenticator.ERROR_AUTH_NEW_USER_MSG.format(email=authid),
                                       email=authid)
            else:
                # send user message with activation code
                await self._messenger.send_user_login_mail(authid, activation.code)
                self.log.info('User login mail sent to {authid} with activation code {activation_code}',
                              authid=hlid(authid),
                              activation_code=hl(activation.code, color='red', bold=True))

                # deny authentication by raising an error and providing feedback to client
                raise ApplicationError(Authenticator.ERROR_AUTH_REGISTERED_USER,
                                       Authenticator.ERROR_AUTH_REGISTERED_USER_MSG.format(email=authid),
                                       email=authid)
        else:

            self.log.debug('Activation found in database:\n{activation}', activation=activation)

            if activation.status == ActivationStatus.ACTIVE:

                # ok, so the user's public key is known and active .. the 99% case

                # user provided an activation code, though there is no activation currently
                if activation_code:
                    raise ApplicationError(Authenticator.ERROR_AUTH_NO_PENDING_ACT,
                                           Authenticator.ERROR_AUTH_NO_PENDING_ACT_MSG)

                # .. if the user wants to join the global users realm, allow that,
                # but ignore any authrole that might have been requested
                if realm is None or realm == Authenticator.GLOBAL_USER_REALM:
                    auth = {
                        'pubkey': pubkey,
                        'realm': Authenticator.GLOBAL_USER_REALM,
                        'authid': authid,
                        'role': Authenticator.GLOBAL_USER_REALM_USER_ROLE,
                        'extra': None,
                        'cache': False
                    }
                    self.log.info('Found user {authid} with active pubkey, authenticating for global user realm',
                                  authid=authid)

                    return auth

                # .. if the user wants to join a specific (management) realm, we need to check more ..
                else:
                    user_roles = None
                    with self._db.begin() as txn:
                        user_oid = self._schema.idx_users_by_email[txn, authid]
                        if user_oid:
                            mrealm_oid = self._schema.idx_mrealms_by_name[txn, realm]
                            if mrealm_oid:
                                user_roles = self._schema.users_mrealm_roles[txn, (user_oid, mrealm_oid)]
                                self.log.info('user roles {user_roles}', user_roles=user_roles)
                            else:
                                self.log.info('no mrealm with name "{realm}"', realm=realm)
                        else:
                            self.log.info('no user for authid "{authid}"', authid=authid)

                    if not user_roles:
                        raise Exception('no realm "{}" or user not permitted'.format(realm))
                    else:
                        self.log.info('user has {roles} roles on realm {realm}', roles=user_roles.roles, realm=realm)

                    if authrole is None:
                        # the user did not request a specific role, so take the first one?
                        # or take the role with the highest privileges? or lowest? FIXME
                        authrole = min(user_roles.roles)  # the minimum is the _highest_ permission (OWNER=1)
                    else:
                        MAP = {
                            'owner': UserRole.OWNER,
                            'admin': UserRole.ADMIN,
                            'user': UserRole.USER,
                            'guest': UserRole.GUEST,
                        }
                        authrole = MAP.get(authrole, None)

                    # the user requested a specific role: check that the role is in the
                    # list of permitted roles the user may take on the management realm
                    if authrole not in user_roles.roles:
                        raise Exception('not authorized for role {}'.format(authrole))

                    # map the integer authrole to a string
                    MAP = {
                        UserRole.OWNER: 'owner',
                        UserRole.ADMIN: 'admin',
                        UserRole.USER: 'user',
                        UserRole.GUEST: 'guest',
                    }
                    authrole = MAP.get(authrole, None)

                    auth = {
                        'pubkey': pubkey,
                        'realm': realm,
                        'authid': authid,
                        'role': authrole,
                        'extra': None,
                        'cache': False
                    }
                    self.log.info('auth=\n{auth}', auth=auth)

                    self.log.info(
                        'Authenticated CF user with pubkey {pubkey}.. as authid "{authid}" on realm "{realm}"',
                        authid=authid,
                        realm=realm,
                        pubkey=pubkey[:16])

                    return auth

            elif activation.status == ActivationStatus.PENDING:

                now = datetime.utcnow()

                passed_secs = (now - activation.created).total_seconds()
                passed_secs_str = humanize.naturaldelta(passed_secs)

                if not activation_code:
                    raise ApplicationError(Authenticator.ERROR_AUTH_PENDING_ACT,
                                           Authenticator.ERROR_AUTH_PENDING_ACT_MSG.format(passed_secs_str))

                if activation_code != activation.code:
                    msg = 'code does not match pending one'
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_ACT_CODE,
                                           Authenticator.ERROR_AUTH_INVALID_ACT_CODE_MSG.format(msg))

                # check if the activation is expired (15min). if so, delete it, and bail out
                if passed_secs > (60 * 15):

                    with self._db.begin(write=True) as txn:
                        oid = self._schema.idx_act_tokens_by_authid_pubkey[txn, authid + pubkey]
                        if oid:
                            del self._schema.activation_tokens[txn, oid]
                        else:
                            raise Exception('no such activation')

                    msg = 'code created {} ago has expired'.format(passed_secs_str)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_ACT_CODE,
                                           Authenticator.ERROR_AUTH_INVALID_ACT_CODE_MSG.format(msg))

                # sanitize the stored activation info against what the client provided
                if activation.atype not in [ActivationType.LOGIN, ActivationType.REGISTRATION]:
                    msg = 'activation type "{}" is not for user login/registration.'.format(activation.atype)
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_ACT_CODE,
                                           Authenticator.ERROR_AUTH_INVALID_ACT_CODE_MSG.format(msg))

                if activation.email != authid:
                    msg = 'email associated with activation code does not match authid provided by client.'
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_ACT_CODE,
                                           Authenticator.ERROR_AUTH_INVALID_ACT_CODE_MSG.format(msg))

                if activation.pubkey != pubkey:
                    msg = 'pubkey associated with activation code does not match pubkey provided by client.'
                    raise ApplicationError(Authenticator.ERROR_AUTH_INVALID_ACT_CODE,
                                           Authenticator.ERROR_AUTH_INVALID_ACT_CODE_MSG.format(msg))

                activation.status = ActivationStatus.ACTIVE
                activation.activated = now

                # create user
                if is_new_user:
                    user = User()
                    user.oid = uuid.uuid4()
                    user.email = authid
                    user.registered = datetime.utcnow()

                    with self._db.begin(write=True) as txn:
                        self._schema.users[txn, user.oid] = user

                    self.log.info('New user stored in database:\n{user}', user=user)
                else:
                    self.log.info('User already stored in database:\n{user}', user=user)

                # update user-pubkey activation
                with self._db.begin(write=True) as txn:
                    oid = self._schema.idx_act_tokens_by_authid_pubkey[txn, authid + pubkey]
                    if oid:
                        self._schema.activation_tokens[txn, oid] = activation
                    else:
                        raise Exception('no such activation')

                # immediately auth user on global users realm
                auth = {
                    'pubkey': pubkey,
                    'realm': Authenticator.GLOBAL_USER_REALM,
                    'authid': authid,
                    'role': Authenticator.GLOBAL_USER_REALM_USER_ROLE,
                    'extra': None,
                    'cache': False
                }

                self.log.info('found principal for public key {pubkey} of {authid}',
                              pubkey=pubkey,
                              authid=auth['authid'])

                return auth

            else:
                raise Exception('internal error: unprocessed activation status {}'.format(activation.status))

    def _send_user_login_mail(self, receiver, activation_code):
        subject = 'Crossbar.io: your LOGIN code'
        text = 'We have received a login request for your account. Please use this activation code: {}'.format(
            activation_code)
        return self._messenger.send_message(receiver, subject, text)

    def _send_user_registration_mail(self, receiver, activation_code):
        subject = 'Crossbar.io: your REGISTRATION code'
        text = 'We have received a registration request for your account. Please use this activation code: {}'.format(
            activation_code)
        return self._messenger.send_message(receiver, subject, text)



# coding=utf8

##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.uri import error


@error('xbr.network.error.username_already_exists')
class UsernameAlreadyExists(ApplicationError):
    """
    An action could not be performed because the chosen username already exists.
    """
    def __init__(self, username, alt_username=None):
        if alt_username:
            msg = 'username "{}" already exists. alternative available username "{}"'.format(username, alt_username)
        else:
            msg = 'username "{}" already exists'.format(username)
        super().__init__('xbr.network.error.username_already_exists', msg, alt_username=alt_username)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.internet.defer import inlineCallbacks

from autobahn.twisted.wamp import ApplicationSession
from autobahn.twisted.util import sleep


class LogTester(ApplicationSession):
    """
    Sample WAMP component to test logging at run-time.
    """
    @inlineCallbacks
    def onJoin(self, details):
        LOG_PREFIX = '*** SAMPLE *** [{}]'.format(details.session)
        config = self.config.extra or {'iterations': 300, 'delay': .2}

        self.log.info('{prefix} joined realm "{realm}"', prefix=LOG_PREFIX, realm=details.realm)
        self.log.info('{prefix} config={config}', prefix=LOG_PREFIX, config=self.config.extra)

        self._tick = 1
        for i in range(config['iterations']):
            self.log.info('{prefix} TICK:', prefix=LOG_PREFIX)
            for fn, lvl in [(self.log.trace, 'TRACE'), (self.log.debug, 'DEBUG'), (self.log.info, 'INFO '),
                            (self.log.warn, 'WARN '), (self.log.error, 'ERROR')]:
                fn('{prefix} {lvl} - TICK {tick}', prefix=LOG_PREFIX, lvl=lvl, tick=self._tick)
            self._tick += 1
            yield sleep(config['delay'])

        self.log.info('{prefix} DONE!', prefix=LOG_PREFIX)

        self.leave()



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from autobahn.util import utcnow

from crossbar.edge.worker.monitor._base import Monitor

__all__ = ('IOMonitor', )


class IOMonitor(Monitor):
    """
    IO monitoring. This is using Linux procfs to get measurements.
    """

    ID = u'diskio'

    def __init__(self, config=None):
        Monitor.__init__(self, config)

        self._storage = self._config.get(u'storage', [])

        # flat list of block devices
        #
        self._devices = []
        for subsystem in self._storage:
            for device, _ in subsystem['devices']:
                self._devices.append(device)

        # map indexed by device holding last raw (cumulative) values
        self._last = {}

        # map indexed by device holding last values (since previous sample)
        self._values = {}

        for device in self._devices:
            self._last[device] = None
            self._values[device] = None

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        Monitor.poll(self)

        # create new, empty event
        #
        current = {
            # the UTC timestamp when measurement was taken
            u'timestamp': utcnow(),

            # the effective last period in secods
            u'last_period': self._last_period,

            # storage subsystem measurements
            u'subsystems': []
        }

        # normalize with effective period
        diff = self._last_period or 1.

        # get IO stats per device from procfs (/sys/block/<device>/stat)
        # see: https://www.kernel.org/doc/Documentation/block/stat.txt
        #
        for device in self._devices:
            with open('/sys/block/{}/stat'.format(device)) as fd:
                res = fd.read()

                new = [int(s.strip()) for s in res.split()]

                if not self._last[device]:
                    self._last[device] = new

                last = self._last[device]

                self._values[device] = {
                    u'read_ios': int((new[0] - last[0]) / diff),
                    u'read_merges': int((new[1] - last[1]) / diff),
                    u'read_bytes': int(512 * (new[2] - last[2]) / diff),
                    u'read_ticks': int((new[3] - last[3]) / diff),
                    u'write_ios': int((new[4] - last[4]) / diff),
                    u'write_merges': int((new[5] - last[5]) / diff),
                    u'write_bytes': int(512 * (new[6] - last[6]) / diff),
                    u'write_ticks': int((new[7] - last[7]) / diff),
                    u'in_flight': new[8],
                    u'io_ticks': int((new[9] - last[9]) / diff),
                    u'time_in_queue': int((new[10] - last[10]) / diff)
                }

                self._last[device] = new

        # transform raw measurements into target event structure
        #
        for subsys in self._storage:
            subsystem = {u'id': subsys[u'id'], u'devices': []}

            for device_id, device_label in subsys[u'devices']:
                values = self._values[device_id]
                device = {
                    u'id': device_id,
                    u'type': device_label,
                    u'read_ios': values['read_ios'],
                    u'read_bytes': values['read_bytes'],
                    u'read_ms': values['read_ticks'],
                    u'write_ios': values['write_ios'],
                    u'write_bytes': values['write_bytes'],
                    u'write_ms': values['write_ticks'],
                    u'in_flight': values['in_flight'],
                    u'active_ms': values['io_ticks'],
                    u'wait_ms': values['time_in_queue'],
                }
                subsystem[u'devices'].append(device)

            current[u'subsystems'].append(subsystem)

        self._last_value = current

        return self._last_value



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteRouterApi', )


class RemoteRouterApi(RemoteApi):

    # remote procedures and topics are exposed under this prefix
    PREFIX = u'crossbarfabriccenter.remote.router.'

    PROCS = {
        # these are worker level procedures
        u'worker': [
            u'get_router_realms',
            u'get_router_realm',
            u'get_router_realm_stats',
            u'get_router_realm_by_name',
            u'start_router_realm',
            u'stop_router_realm',
            u'get_router_realm_roles',
            u'get_router_realm_role',
            u'start_router_realm_role',
            u'stop_router_realm_role',
            u'get_router_transports',
            u'get_router_transport',
            u'start_router_transport',
            u'stop_router_transport',
            u'get_web_transport_services',
            u'get_web_transport_service',
            u'start_web_transport_service',
            u'stop_web_transport_service',
            u'get_router_components',
            u'get_router_component',
            u'start_router_component',
            u'stop_router_component',
            u'get_router_realm_links',
            u'get_router_realm_link',
            u'start_router_realm_link',
            u'stop_router_realm_link',
        ],
    }

    EVENTS = {
        # these are worker level topics
        u'worker': [
            u'on_router_realm_starting',
            u'on_router_realm_started',
            u'on_router_realm_stopping',
            u'on_router_realm_stopped',
            u'on_router_realm_role_starting',
            u'on_router_realm_role_started',
            u'on_router_realm_role_stopping',
            u'on_router_realm_role_stopped',
            u'on_router_transport_starting',
            u'on_router_transport_started',
            u'on_router_transport_stopping',
            u'on_router_transport_stopped',
            u'on_web_transport_service_starting',
            u'on_web_transport_service_started',
            u'on_web_transport_service_stopping',
            u'on_web_transport_service_stopped',
            u'on_router_component_starting',
            u'on_router_component_started',
            u'on_router_component_stopping',
            u'on_router_component_stopped',
        ]
    }



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import argparse
import json
import pprint
import os

import docutils
import jinja2
from jinja2 import Environment, FileSystemLoader

from crossbar.shell.util import hl


def rst_filter(rst):
    html = docutils.core.publish_parts(source=rst, writer_name='html')['html_body']
    return jinja2.Markup(html)


def process(schema, template_paths=['templates', 'tests/idl']):
    # http://jinja.pocoo.org/docs/latest/api/#loaders
    loader = FileSystemLoader(template_paths, encoding='utf-8', followlinks=False)

    env = Environment(loader=loader, autoescape=True)
    env.filters['rst'] = rst_filter

    tmpl = env.get_template('service.py')

    print(tmpl)

    contents = tmpl.render(schema=schema)

    print(contents)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('infile', help='FlatBuffers JSON schema input file (.json)')
    parser.add_argument('-t', '--templates', help='Templates folder')
    parser.add_argument('-v', '--verbose', action='store_true', help='Enable verbose processing output.')
    parser.add_argument('-d', '--debug', action='store_true', help='Enable debug output.')

    options = parser.parse_args()

    import txaio
    txaio.use_twisted()

    log = txaio.make_logger()
    txaio.start_logging(level='debug' if options.debug else 'info')

    infile_path = os.path.abspath(options.infile)
    with open(infile_path, 'rb') as f:
        buf = f.read()

    log.info('Loading FlatBuffers JSON schema ({} bytes) ...'.format(len(buf)))

    try:
        schema = json.loads(buf, encoding='utf8')
    except Exception as e:
        log.error(e)

    if options.verbose:
        log.info('Schema metadata:')
        schema_meta_str = pprint.pformat(schema['meta'])
        # log.info(schema_meta_str)
        # log.info('{}'.format(schema_meta_str))
        print(schema_meta_str)

        for o in schema['types'].values():
            if o['type'] == 'interface':
                log.info('interface: {}'.format(hl(o['name'], bold=True)))
                for s in o['slots'].values():
                    log.info('{:>12}: {}'.format(s['type'], hl(s['name'])))

    process(schema)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################
import six

from crossbar._util import hl


def hlval(val):
    return hl('{}'.format(val), color='white', bold=True)


def hlcontract(oid):
    if not isinstance(oid, six.text_type):
        oid = '{}'.format(oid)
    return hl('<{}>'.format(oid), color='magenta', bold=True)



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import json
import treq
import os

from txaio import make_logger

from twisted.internet.defer import inlineCallbacks, returnValue

__all__ = ('Messenger', )


class Messenger(object):

    log = make_logger()

    def __init__(self, submit_url, access_key):
        self._submit_url = submit_url
        self._access_key = access_key
        self._mailgun_from = os.environ.get('MAILGUN_FROM', "Crossbar.io <no-reply@mailing.crossbar.io>")

    def send_user_login_mail(self, receiver, activation_code):
        subject = u'Crossbar.io: your LOGIN code'
        text = u'''Hello from Crossbar.io!

        We have received a login request for your account.

        Please use this activation code: {}

        If you did not request a login code, then just ignore this email.
        '''.format(activation_code)
        return self.send_message(receiver, subject, text)

    def send_user_registration_mail(self, receiver, activation_code):
        subject = u'Crossbar.io: your REGISTRATION code'
        text = u'''Hello from Crossbar.io!

        We have received a registration request for this email address.

        Please use this activation code:
            {}
        e.g. to authenticate using the Crossbar.io shell do
            cbsh auth --code {}

        If you did not request a registration, then just ignore this email. No Crossbar.io account will be created for your email address!
        '''.format(activation_code, activation_code)
        return self.send_message(receiver, subject, text)

    @inlineCallbacks
    def send_message(self, receiver, subject, text):

        self.log.info('sending mail via mailgun: receiver={receiver}, subject="{subject}", textlen={textlen}',
                      receiver=receiver,
                      subject=subject,
                      textlen=len(text))

        data = {"from": self._mailgun_from, "to": [receiver], "subject": subject, "text": text}

        res = None
        self.log.debug('Mailgun URL={url}', url=self._submit_url)
        try:
            if self._access_key and self._submit_url:
                res = yield treq.post(self._submit_url, auth=("api", self._access_key), data=data)
            else:
                self.log.warn('Mailgun not configured! This is the mail that would have been sent: {mail}',
                              mail=json.dumps(data))
                res = None
        except Exception as e:
            print('Exception:', e)
            print('Result:', res)
            self.log.failure()
            raise

        returnValue(res)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import time
import datetime
import psutil

from twisted.internet.threads import deferToThread

from autobahn.util import utcnow, utcstr

from crossbar.edge.worker.monitor._base import Monitor


class ProcessMonitor(Monitor):

    ID = u'process'

    def __init__(self, config=None):
        Monitor.__init__(self, config)

        # Map of process types.
        self._ptypes = self._config.get(u'process_types', {})

        # Process type to map otherwise unmapped processes to.
        self._ptype_unassigned = self._config.get(u'process_type_unassigned', u'unassigned')

        # Filter processes by these types.
        filter_ptypes = self._config.get(u'filter_process_types', None)
        self._filter_ptypes = set(filter_ptypes) if filter_ptypes else None

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        Monitor.poll(self)
        return deferToThread(self._poll)

    def _poll(self):

        # create new, empty event
        #
        current = {
            # the UTC timestamp when measurement was taken
            u'timestamp': utcnow(),

            # the effective last period in secods
            u'last_period': self._last_period,

            # actual process statistics
            u'processes': {}
        }

        # normalize with effective period
        diff = self._last_period or 1.

        cmd_started = time.time()

        for proc in psutil.process_iter():
            pinfo = {}
            try:
                # map process executable to (user defined) process type
                #
                proc_id = hash(proc) % (2**53)  # JavaScript represent numbers as IEEE double!
                pinfo[u'exe'] = proc.exe()
                pinfo[u'type'] = self._ptypes.get(pinfo[u'exe'], self._ptype_unassigned)

                # if no filtering by process type is active OR if the process type is in
                # the filter list, do the thing ..
                #
                if self._filter_ptypes is None or pinfo[u'type'] in self._filter_ptypes:

                    pinfo[u'id'] = proc_id
                    pinfo[u'pid'] = proc.pid
                    pinfo[u'user'] = proc.username()
                    pinfo[u'status'] = proc.status()
                    pinfo[u'name'] = proc.name()
                    pinfo[u'cmdline'] = u' '.join(proc.cmdline())

                    created = proc.create_time()
                    pinfo[u'created'] = utcstr(datetime.datetime.fromtimestamp(created))

                    pinfo[u'num_fds'] = proc.num_fds()
                    pinfo[u'num_threads'] = proc.num_threads()
                    pinfo[u'num_fds'] = proc.num_fds()

                    # the following values are cumulative since process creation!
                    #
                    num_ctx_switches = proc.num_ctx_switches()
                    pinfo[u'num_ctx_switches_voluntary'] = num_ctx_switches.voluntary
                    pinfo[u'num_ctx_switches_involuntary'] = num_ctx_switches.involuntary

                    iocounters = proc.io_counters()
                    pinfo[u'read_ios'] = iocounters.read_count
                    pinfo[u'write_ios'] = iocounters.write_count
                    pinfo[u'read_bytes'] = iocounters.read_bytes
                    pinfo[u'write_bytes'] = iocounters.write_bytes

                    cpu = proc.cpu_times()
                    pinfo[u'cpu_user'] = cpu.user
                    pinfo[u'cpu_system'] = cpu.system

                    current[u'processes'][proc_id] = pinfo

            except psutil.NoSuchProcess:
                pass

        current[u'command_duration'] = time.time() - cmd_started

        if self._last_value:
            for proc_id in current[u'processes']:
                if proc_id in self._last_value[u'processes']:
                    proc = current[u'processes'][proc_id]
                    last = self._last_value[u'processes'][proc_id]
                    for key in [
                            u'read_ios', u'write_ios', u'read_bytes', u'write_bytes', u'cpu_user', u'cpu_system',
                            u'num_ctx_switches_voluntary', u'num_ctx_switches_involuntary'
                    ]:
                        proc[u'{}_per_sec'.format(key)] = float(proc[key] - last[key]) / diff
                else:
                    # should not arrive here!
                    pass
            self._last_value = current
            return current
        else:
            self._last_value = current
            return None



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################
import os
import socket

from twisted.internet.defer import inlineCallbacks, Deferred, returnValue, gatherResults
from twisted.internet.defer import succeed

from txaio import make_logger

from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.cryptosign import SigningKey
from autobahn.wamp.types import CallOptions, ComponentConfig

from crossbar._util import hltype, hlid, hluserid, hl

from crossbar.router.router import RouterFactory
from crossbar.router.session import RouterSessionFactory
from crossbar.router.service import RouterServiceAgent
from crossbar.worker.types import RouterRealm
from crossbar.common.checkconfig import NODE_SHUTDOWN_ON_WORKER_EXIT
from crossbar.common.key import _maybe_generate_node_key
from crossbar.node.controller import NodeController


class NodeOptions(object):
    def __init__(self, debug_lifecycle=False, debug_programflow=False, enable_vmprof=False):

        self.debug_lifecycle = debug_lifecycle
        self.debug_programflow = debug_programflow
        self.enable_vmprof = enable_vmprof


class Node(object):
    """
    Crossbar.io Standalone node personality.
    """
    NODE_CONTROLLER = NodeController

    ROUTER_SERVICE = RouterServiceAgent

    CONFIG_SOURCE_EMPTY = 0
    CONFIG_SOURCE_DEFAULT = 1
    CONFIG_SOURCE_LOCALFILE = 2
    CONFIG_SOURCE_XBRNETWORK = 4
    CONFIG_SOURCE_TO_STR = {
        0: 'empty',
        1: 'default',
        2: 'localfile',
        4: 'xbrnetwork',
    }

    # A Crossbar.io node is the running a controller process and one or multiple
    # worker processes.
    # A single Crossbar.io node runs exactly one instance of this class, hence
    # this class can be considered a system singleton.

    log = make_logger()

    def __init__(self, personality, cbdir=None, reactor=None, native_workers=None, options=None):
        """

        :param cbdir: The node directory to run from.
        :type cbdir: unicode
        :param reactor: Reactor to run on.
        :type reactor: :class:`twisted.internet.reactor` or None
        """
        self.personality = personality
        self.options = options or NodeOptions()

        self._native_workers = personality.native_workers

        # node directory
        self._cbdir = cbdir or '.'

        # reactor we should run on
        if reactor is None:
            from twisted.internet import reactor
        self._reactor = reactor

        # allow overriding to add (or remove) native-worker types
        if native_workers is not None:
            self._native_workers = native_workers

        # local node management router
        self._router_factory = None

        # session factory for node management router
        self._router_session_factory = None

        # the node controller realm
        self._realm = 'crossbar'

        # config of this node
        self._config = None

        # source(s) of the config of this node
        self._config_source = 0

        # node private key :class:`autobahn.wamp.cryptosign.SigningKey`
        self._node_key = None

        # when running in managed mode, this will hold the session to CFC
        self._manager = None

        # the node's management realm when running in managed mode (this comes from CFC!)
        self._management_realm = None

        # the node's ID when running in managed mode (this comes from CFC!)
        self._node_id = None

        # node extra when running in managed mode (this comes from CFC!)
        self._node_extra = None

        # node controller session (a singleton ApplicationSession embedded
        # in the local node router)
        self._controller = None
        self._service_sessions = {}

        # node shutdown triggers, one or more of checkconfig.NODE_SHUTDOWN_MODES
        self._node_shutdown_triggers = [NODE_SHUTDOWN_ON_WORKER_EXIT]

        # will be filled with a Deferred in start(). the Deferred will fire when
        # the node has shut down, and the result signals if shutdown was clean
        self._shutdown_complete = None

        # wil ve filled with a Deferred in start(). the Deferred will fire when
        # the node has booted completely
        self._boot_complete = None

        # for node elements started under specific IDs, and where
        # the node configuration does not specify an ID, use a generic
        # name numbered sequentially using the counters here
        self._worker_no = 1
        self._realm_no = 1
        self._role_no = 1
        self._connection_no = 1
        self._transport_no = 1
        self._webservice_no = 1
        self._component_no = 1

    @property
    def realm(self):
        return self._realm

    @property
    def key(self):
        """
        Returns the node (private signing) key pair.

        :return: The node key.
        :rtype: :class:`autobahn.wamp.cryptosign.SigningKey`
        """
        return self._node_key

    def load_keys(self, cbdir):
        """
        Load node public-private key pair from key files, possibly generating a new key pair if
        none exists.

        This is the _first_ function being called after the Node has been instantiated.

        IMPORTANT: this function is run _before_ start of Twisted reactor!
        """
        was_new, self._node_key = _maybe_generate_node_key(cbdir)
        return was_new

    def load_config(self, configfile=None, default=None):
        """
        Check and load the node configuration from:

        * from ``.crossbar/config.json`` or
        * from built-in (empty) default configuration

        This is the _second_ function being called after the Node has been instantiated.

        IMPORTANT: this function is run _before_ start of Twisted reactor!
        """
        self.log.debug('{klass}.load_config(configfile={configfile}, default={default}) ..',
                       klass=self.__class__.__name__,
                       configfile=configfile,
                       default=default)
        if configfile:
            config_path = os.path.abspath(os.path.join(self._cbdir, configfile))

            # the following will read the config, check the config and replace
            # environment variable references in configuration values ("${MYVAR}") and
            # finally return the parsed configuration object
            self._config = self.personality.check_config_file(self.personality, config_path)
            config_source = Node.CONFIG_SOURCE_LOCALFILE
        else:
            config_path = None
            if default:
                self._config = default
                config_source = Node.CONFIG_SOURCE_DEFAULT
            else:
                self._config = {'version': 2, 'controller': {}, 'workers': []}
                config_source = Node.CONFIG_SOURCE_EMPTY

            self.personality.check_config(self.personality, self._config)

        if 'controller' in self._config and 'keyring' in self._config['controller']:
            keyring_type = self._config['controller']['keyring']['type']
            if keyring_type == 'file':
                key_path = self._config['controller']['keyring']['path']
                self._node_key = SigningKey.from_raw_key(key_path)
            else:
                raise RuntimeError("NotImplemented: hsm authentication is currently not implemented.")

        return config_source, config_path

    def is_key_loaded(self):
        return self._node_key is not None

    def _add_global_roles(self):
        controller_role_config = {
            # there is exactly 1 WAMP component authenticated under authrole "controller": the node controller
            "name":
            "controller",
            "permissions": [{
                # the node controller can (locally) do "anything"
                "uri": "crossbar.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": True,
                    "publish": True,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }]
        }
        self._router_factory.add_role(self._realm, controller_role_config)
        self.log.info('{func} node-wide role "{authrole}" added on node management router realm "{realm}"',
                      func=hltype(self._add_global_roles),
                      authrole=hlid(controller_role_config['name']),
                      realm=hlid(self._realm))

    def _add_worker_role(self, worker_auth_role, options):
        worker_role_config = {
            # each (native) worker is authenticated under a worker-specific authrole
            "name":
            worker_auth_role,
            "permissions": [
                # the worker requires these permissions to work:
                {
                    # management API provided by the worker. note that the worker management API is provided under
                    # the URI prefix "crossbar.worker.<worker_id>". note that the worker is also authenticated
                    # under authrole <worker_auth_role> on realm "crossbar"
                    "uri": worker_auth_role,
                    "match": "prefix",
                    "allow": {
                        "call": True,
                        "register": True,
                        "publish": True,
                        "subscribe": True
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                },
                {
                    # controller procedure called by the worker (to check for controller status)
                    "uri": "crossbar.get_status",
                    "match": "exact",
                    "allow": {
                        "call": True,
                        "register": False,
                        "publish": False,
                        "subscribe": False
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                }
            ]
        }
        # if configured to expose the controller connection within the worker (to make it available
        # in user code such as dynamic authenticators and router/container components), also add
        # permissions to actually use the (local) node management API
        if options.get('expose_controller', True):
            vendor_permissions = {
                "uri": "crossbar.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": False,
                    "publish": False,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }
            worker_role_config["permissions"].append(vendor_permissions)

        self._router_factory.add_role(self._realm, worker_role_config)

        self.log.info('worker-specific role "{authrole}" added on node management router realm "{realm}" {func}',
                      func=hltype(self._add_worker_role),
                      authrole=hlid(worker_role_config['name']),
                      realm=hlid(self._realm))

    def _drop_worker_role(self, worker_auth_role):
        self._router_factory.drop_role(self._realm, worker_auth_role)

    def _extend_worker_args(self, args, options):
        pass

    def _add_extra_controller_components(self, controller_config):
        pass

    def _set_shutdown_triggers(self, controller_options):
        # allow to override node shutdown triggers
        #
        if 'shutdown' in controller_options:
            self._node_shutdown_triggers = controller_options['shutdown']
            self.log.info("Using node shutdown triggers {triggers} from configuration",
                          triggers=self._node_shutdown_triggers)
        else:
            self._node_shutdown_triggers = [NODE_SHUTDOWN_ON_WORKER_EXIT]
            self.log.info("Using default node shutdown triggers {triggers}", triggers=self._node_shutdown_triggers)

    def set_service_session(self, session, realm, authrole=None):
        self.log.info('{func}(session={session}, realm="{realm}", authrole="{authrole}")',
                      func=hltype(self.set_service_session),
                      session=session,
                      realm=hlid(realm),
                      authrole=hlid(authrole))
        if realm not in self._service_sessions:
            self._service_sessions[realm] = {}
        self._service_sessions[realm][authrole] = session

    def get_service_session(self, realm, authrole=None):
        if realm in self._service_sessions:
            if authrole in self._service_sessions[realm]:
                session = self._service_sessions[realm][authrole]
                self.log.info('{func}(session={session}, realm="{realm}", authrole="{authrole}")',
                              func=hltype(self.get_service_session),
                              session=session,
                              realm=hlid(realm),
                              authrole=hlid(authrole))
                return succeed(session)
        return succeed(None)

    def stop(self, restart=False):
        self._controller._shutdown_was_clean = True
        return self._controller.shutdown(restart=restart)

    @inlineCallbacks
    def start(self, node_id=None):
        """
        Starts this node. This will start a node controller and then spawn new worker
        processes as needed.

        The node keys (``load_keys``) and configuration (``load_config``) has to be loaded
        before starting the node.

        This is the _third_ function being called after the Node has been instantiated.
        """
        self.log.info('{note} [{method}]',
                      note=hl('Starting node ..', color='green', bold=True),
                      method=hltype(Node.start))

        # a configuration must have been loaded before
        if not self._config:
            self.log.warn('no node configuration set - will use empty node configuration!')
            self._config = {}

        # a node can only be started once for now
        assert self._shutdown_complete is None
        assert self._node_id is None

        # get controller config/options
        controller_config = self._config.get('controller', {})
        controller_options = controller_config.get('options', {})

        # the node ID: CLI takes precedence over config over hostname
        if node_id:
            self._node_id = node_id
            _node_id_source = 'explicit run-time argument'
        elif 'id' in controller_config:
            self._node_id = controller_config['id']
            _node_id_source = 'explicit configuration'
        else:
            self._node_id = '{}-{}'.format(socket.gethostname(), os.getpid()).lower()
            _node_id_source = 'hostname/pid'
        self.log.info('Node ID {node_id} set from {node_id_source}',
                      node_id=hlid(self._node_id),
                      node_id_source=_node_id_source)

        # set controller process title
        try:
            import setproctitle
        except ImportError:
            self.log.warn("Warning, could not set process title (setproctitle not installed)")
        else:
            setproctitle.setproctitle(controller_options.get('title', 'crossbar-controller'))

        # add the node controller singleton component
        self._controller = self.NODE_CONTROLLER(self)

        # local node management router
        self._router_factory = RouterFactory(self._node_id, None, None)
        self._router_session_factory = RouterSessionFactory(self._router_factory)

        # start node-wide realm on node management router
        rlm_config = {'name': self._realm}
        rlm = RouterRealm(self._controller, None, rlm_config)
        router = self._router_factory.start_realm(rlm)

        # setup global static roles
        self._add_global_roles()

        # always add a realm service session
        cfg = ComponentConfig(self._realm, controller=self._controller)
        rlm.session = (self.ROUTER_SERVICE)(cfg, router)
        self._router_session_factory.add(rlm.session, router, authid='serviceagent', authrole='trusted')
        self.log.info('{func} router service agent session attached [{router_service}]',
                      func=hltype(self.start),
                      router_service=hltype(self.ROUTER_SERVICE))

        self._router_session_factory.add(self._controller, router, authid='nodecontroller', authrole='controller')
        self._service_sessions[self._realm] = self._controller
        self.log.info('{func} node controller session attached [{node_controller}]',
                      func=hltype(self.start),
                      node_controller=hltype(self.NODE_CONTROLLER))

        # add extra node controller components
        self._add_extra_controller_components(controller_config)

        # setup Node shutdown triggers
        self._set_shutdown_triggers(controller_options)

        # setup node shutdown Deferred
        self._shutdown_complete = Deferred()

        # setup node booted complete Deferred
        self._boot_complete = Deferred()

        # startup the node personality ..
        self.log.info('{func}::NODE_BOOT_BEGIN[node_id="{node_id}"]',
                      node_id=hlid(self._node_id),
                      func=hltype(self.personality.Node.boot))
        res = yield self.personality.Node.boot(self)
        self.log.info('{func}::NODE_BOOT_COMPLETE[node_id="{node_id}"]',
                      func=hltype(self.personality.Node.boot),
                      node_id=hlid(self._node_id))

        # notify observers of boot completition
        self._boot_complete.callback(res)

        # notify systemd that we are fully up and running
        try:
            import sdnotify
        except ImportError:
            # do nothing on non-systemd platforms
            pass
        else:
            sdnotify.SystemdNotifier().notify("READY=1")

        # return a shutdown deferred which we will fire to notify the code that
        # called start() - which is the main crossbar boot code
        res = {'shutdown_complete': self._shutdown_complete}
        returnValue(res)

    def boot(self):
        self.log.info('Booting node {method}', method=hltype(Node.boot))
        return self.boot_from_config(self._config)

    @inlineCallbacks
    def boot_from_config(self, config):
        """
        Startup elements in the node as specified in the provided node configuration.
        """
        # get controller configuration subpart
        controller = config.get('controller', {})
        parallel_worker_start = controller.get('options', {}).get('enable_parallel_worker_start', False)

        self.log.info(
            '{bootmsg} {method}',
            bootmsg=hl(
                'Booting node from local configuration [parallel_worker_start={}] ..'.format(parallel_worker_start),
                color='green',
                bold=True),
            method=hltype(Node.boot_from_config))

        # start Manhole in node controller
        if 'manhole' in controller:
            yield self._controller.call('crossbar.start_manhole', controller['manhole'], options=CallOptions())
            self.log.debug("controller: manhole started")

        # startup all workers
        workers = config.get('workers', [])
        if len(workers):
            self.log.info(
                hl('Will start {} worker{} ..'.format(len(workers), 's' if len(workers) > 1 else ''),
                   color='green',
                   bold=True))
        else:
            self.log.info(hl('No workers configured, nothing to do', color='green', bold=True))

        dl = []
        for worker in workers:

            # worker ID
            if 'id' in worker:
                worker_id = worker['id']
            else:
                worker_id = 'worker{:03d}'.format(self._worker_no)
                worker['id'] = worker_id
                self._worker_no += 1

            # worker type: either a native worker ('router', 'container', ..), or a guest worker ('guest')
            worker_type = worker['type']

            # native worker processes setup
            if worker_type in self._native_workers:

                # set logname depending on native worker type
                worker_logname = '{} {}'.format(self._native_workers[worker_type]['logname'], hlid(worker_id))

                # any worker specific options
                worker_options = worker.get('options', {})

                worker_disabled = worker_options.get('disabled', False)

                if worker_disabled:
                    self.log.warn(
                        'SKIP STARTING OF WORKER ! ("{worker_logname}" disabled from config)',
                        worker_logname=worker_logname,
                    )
                else:
                    # start the (native) worker
                    self.log.info(
                        'Order node to start "{worker_logname}" ..',
                        worker_logname=hlid(worker_logname),
                    )

                    d = self._controller.call('crossbar.start_worker',
                                              worker_id,
                                              worker_type,
                                              worker_options,
                                              options=CallOptions())

                    @inlineCallbacks
                    def configure_worker(res, worker_logname, worker_type, worker_id, worker):
                        self.log.info(
                            "Ok, node has started {worker_logname}",
                            worker_logname=worker_logname,
                        )

                        # now configure the worker
                        self.log.info(
                            "Configuring {worker_logname} ..",
                            worker_logname=worker_logname,
                        )

                        # _configure_native_worker_router, _configure_native_worker_container, ...
                        method_name = '_configure_native_worker_{}'.format(worker_type.replace('-', '_'))
                        try:
                            config_fn = getattr(self, method_name)
                        except AttributeError:
                            raise ValueError("A native worker of type '{}' is configured but "
                                             "there is no method '{}' on {}".format(
                                                 worker_type, method_name, type(self)))
                        try:
                            yield config_fn(worker_logname, worker_id, worker)
                        except ApplicationError as e:
                            if e.error != 'wamp.error.canceled':
                                raise

                        self.log.info(
                            'Ok, worker "{worker_logname}" configured and ready!',
                            worker_logname=hlid(worker_logname),
                        )

                    d.addCallback(configure_worker, worker_logname, worker_type, worker_id, worker)

            # guest worker processes setup
            elif worker_type == 'guest':

                # now actually start the (guest) worker ..

                # FIXME: start_worker() takes the whole configuration item for guest workers, whereas native workers
                # only take the options (which is part of the whole config item for the worker)
                d = self._controller.call('crossbar.start_worker',
                                          worker_id,
                                          worker_type,
                                          worker,
                                          options=CallOptions())

            else:
                raise Exception('logic error: unexpected worker_type="{}"'.format(worker_type))

            if parallel_worker_start:
                dl.append(d)
            else:
                yield d

        yield gatherResults(dl)

        self.log.info(hl('Ok, local node configuration ran successfully.', color='green', bold=True))

    @inlineCallbacks
    def _configure_native_worker_common(self, worker_logname, worker_id, worker):
        # expanding PYTHONPATH of the newly started worker is now done
        # directly in NodeController._start_native_worker
        worker_options = worker.get('options', {})
        if False:
            if 'pythonpath' in worker_options:
                added_paths = yield self._controller.call('crossbar.worker.{}.add_pythonpath'.format(worker_id),
                                                          worker_options['pythonpath'],
                                                          options=CallOptions())
                self.log.warn("{worker}: PYTHONPATH extended for {paths}", worker=worker_logname, paths=added_paths)

        # FIXME: as the CPU affinity is in the worker options, this _also_ (see above fix)
        # should be done directly in NodeController._start_native_worker
        if True:
            if 'cpu_affinity' in worker_options:
                new_affinity = yield self._controller.call('crossbar.worker.{}.set_cpu_affinity'.format(worker_id),
                                                           worker_options['cpu_affinity'],
                                                           options=CallOptions())
                self.log.debug("{worker}: CPU affinity set to {affinity}",
                               worker=worker_logname,
                               affinity=new_affinity)

        # this is fine to start after the worker has been started, as manhole is
        # CB developer/support feature anyways (like a vendor diagnostics port)
        if 'manhole' in worker:
            yield self._controller.call('crossbar.worker.{}.start_manhole'.format(worker_id),
                                        worker['manhole'],
                                        options=CallOptions())
            self.log.debug("{worker}: manhole started", worker=worker_logname)

    @inlineCallbacks
    def _configure_native_worker_router(self, worker_logname, worker_id, worker):
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # start realms on router
        for realm in worker.get('realms', []):

            # start realm
            if 'id' in realm:
                realm_id = realm['id']
            else:
                realm_id = 'realm{:03d}'.format(self._realm_no)
                realm['id'] = realm_id
                self._realm_no += 1

            self.log.info(
                "Order {worker_logname} to start Realm {realm_id}",
                worker_logname=worker_logname,
                realm_id=hlid(realm_id),
            )

            yield self._controller.call('crossbar.worker.{}.start_router_realm'.format(worker_id),
                                        realm_id,
                                        realm,
                                        options=CallOptions())

            self.log.info(
                "Ok, {worker_logname} has started Realm {realm_id}",
                worker_logname=worker_logname,
                realm_id=hlid(realm_id),
            )

            # add roles to realm
            for role in realm.get('roles', []):
                if 'id' in role:
                    role_id = role['id']
                else:
                    role_id = 'role{:03d}'.format(self._role_no)
                    role['id'] = role_id
                    self._role_no += 1

                self.log.info(
                    "Order Realm {realm_id} to start Role {role_id}",
                    realm_id=hlid(realm_id),
                    role_id=hlid(role_id),
                )

                yield self._controller.call('crossbar.worker.{}.start_router_realm_role'.format(worker_id),
                                            realm_id,
                                            role_id,
                                            role,
                                            options=CallOptions())

                self.log.info(
                    "Ok, Realm {realm_id} has started Role {role_id}",
                    realm_id=hlid(realm_id),
                    role_id=hlid(role_id),
                )

        # start components to run embedded in the router
        for component in worker.get('components', []):

            if 'id' in component:
                component_id = component['id']
            else:
                component_id = 'component{:03d}'.format(self._component_no)
                component['id'] = component_id
                self._component_no += 1

            yield self._controller.call('crossbar.worker.{}.start_router_component'.format(worker_id),
                                        component_id,
                                        component,
                                        options=CallOptions())
            self.log.info(
                "{logname}: component '{component}' started",
                logname=worker_logname,
                component=component_id,
            )

        # start transports on router
        for transport in worker.get('transports', []):

            if 'id' in transport:
                transport_id = transport['id']
            else:
                transport_id = 'transport{:03d}'.format(self._transport_no)
                transport['id'] = transport_id
                self._transport_no += 1

            add_paths_on_transport_create = False

            self.log.info(
                "Order {worker_logname} to start Transport {transport_id}",
                worker_logname=worker_logname,
                transport_id=hlid(transport_id),
            )

            yield self._controller.call('crossbar.worker.{}.start_router_transport'.format(worker_id),
                                        transport_id,
                                        transport,
                                        create_paths=add_paths_on_transport_create,
                                        options=CallOptions())
            self.log.info(
                "Ok, {worker_logname} has started Transport {transport_id}",
                worker_logname=worker_logname,
                transport_id=hlid(transport_id),
            )

            if not add_paths_on_transport_create:

                if transport['type'] == 'web':
                    paths = transport.get('paths', {})
                elif transport['type'] in ('universal'):
                    paths = transport.get('web', {}).get('paths', {})
                else:
                    paths = None

                # Web service paths
                if paths:
                    for path in sorted(paths):
                        if path != '/':
                            webservice = paths[path]

                            if 'id' in webservice:
                                webservice_id = webservice['id']
                            else:
                                webservice_id = 'webservice{:03d}'.format(self._webservice_no)
                                webservice['id'] = webservice_id
                                self._webservice_no += 1

                            self.log.info(
                                "Order Transport {transport_id} to start Web Service {webservice_id}",
                                transport_id=hlid(transport_id),
                                webservice_id=hlid(webservice_id),
                                path=hluserid(path),
                            )

                            yield self._controller.call(
                                'crossbar.worker.{}.start_web_transport_service'.format(worker_id),
                                transport_id,
                                path,
                                webservice,
                                options=CallOptions())
                            self.log.info(
                                "Ok, Transport {transport_id} has started Web Service {webservice_id}",
                                transport_id=hlid(transport_id),
                                webservice_id=hlid(webservice_id),
                                path=hluserid(path),
                            )

        # start rlinks for realms
        dl = []
        for realm in worker.get('realms', []):
            realm_id = realm['id']
            for i, rlink in enumerate(realm.get('rlinks', [])):
                if 'id' in rlink:
                    rlink_id = rlink['id']
                else:
                    rlink_id = 'rlink{:03d}'.format(i)
                    rlink['id'] = rlink_id

                self.log.info(
                    'Starting router-to-router "{rlink_id}" on realm "{realm_id}" ..',
                    realm_id=hlid(realm_id),
                    rlink_id=hlid(rlink_id),
                )

                d = self._controller.call('crossbar.worker.{}.start_router_realm_link'.format(worker_id),
                                          realm_id,
                                          rlink_id,
                                          rlink,
                                          options=CallOptions())

                def done(_):
                    self.log.info(
                        'Ok, router-to-router {rlink_id} started on realm "{realm_id}".',
                        realm_id=hlid(realm_id),
                        rlink_id=hlid(rlink_id),
                    )

                d.addCallback(done)
                dl.append(d)

        # FIXME: rlinks must be started without waiting for them to be established. otherwise the start of other stuff
        # is waiting for all rlinks to be up!
        d = gatherResults(dl)

    @inlineCallbacks
    def _configure_native_worker_container(self, worker_logname, worker_id, worker):
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # start components to run embedded in the container
        #
        for component in worker.get('components', []):

            if 'id' in component:
                component_id = component['id']
            else:
                component_id = 'component{:03d}'.format(self._component_no)
                component['id'] = component_id
                self._component_no += 1

            yield self._controller.call('crossbar.worker.{}.start_component'.format(worker_id),
                                        component_id,
                                        component,
                                        options=CallOptions())
            self.log.info("{worker}: component '{component_id}' started",
                          worker=worker_logname,
                          component_id=component_id)

    @inlineCallbacks
    def _configure_native_worker_websocket_testee(self, worker_logname, worker_id, worker):
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # start transport on websocket-testee
        transport = worker['transport']

        transport_id = 'transport{:03d}'.format(self._transport_no)
        transport['id'] = transport_id
        self._transport_no = 1

        yield self._controller.call('crossbar.worker.{}.start_websocket_testee_transport'.format(worker_id),
                                    transport_id,
                                    transport,
                                    options=CallOptions())
        self.log.info(
            "{logname}: transport '{tid}' started",
            logname=worker_logname,
            tid=transport_id,
        )

    @inlineCallbacks
    def _configure_native_worker_proxy(self, worker_logname, worker_id, worker):
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # set up backend connections on the proxy

        for i, connection_name in enumerate(worker.get('connections', {})):
            self.log.debug(
                "Starting connection {index}: {name}",
                index=i,
                name=connection_name,
            )
            yield self._controller.call(
                'crossbar.worker.{}.start_proxy_connection'.format(worker_id),
                connection_name,
                worker['connections'].get(connection_name, {}),
            )

        # set up realms and roles on the proxy

        for i, realm_name in enumerate(worker.get('routes', {})):
            roles = worker['routes'][realm_name]
            for role_id, connections in roles.items():
                if not isinstance(connections, list):
                    connections = [connections]  # used to be a single string, now a list of strings
                for connection_id in connections:
                    self.log.debug(
                        "Starting proxy realm route {realm}, {role} to {connection}",
                        realm=realm_name,
                        role=role_id,
                        connection=connection_id,
                    )
                    yield self._controller.call(
                        'crossbar.worker.{}.start_proxy_realm_route'.format(worker_id),
                        realm_name,
                        {role_id: connection_id},
                    )

        # start transports on proxy

        for i, transport in enumerate(worker.get('transports', [])):

            if 'id' in transport:
                transport_id = transport['id']
            else:
                transport_id = 'transport{:03d}'.format(i)
                transport['id'] = transport_id

            self.log.info(
                "Order {worker_logname} to start Transport {transport_id}",
                worker_logname=worker_logname,
                transport_id=hlid(transport_id),
            )

            # XXX we're doing startup, and beginning proxy workers --
            # want to share the web-transport etc etc stuff between
            # these and other kinds of routers / transports
            yield self._controller.call('crossbar.worker.{}.start_proxy_transport'.format(worker_id),
                                        transport_id,
                                        transport,
                                        options=CallOptions())

            if transport['type'] == 'web':
                paths = transport.get('paths', {})
            elif transport['type'] == 'universal':
                paths = transport.get('web', {}).get('paths', {})
            else:
                paths = None

            # Web service paths
            if paths:
                for path in sorted(paths):
                    if path != '/':
                        webservice = paths[path]

                        if 'id' in webservice:
                            webservice_id = webservice['id']
                        else:
                            webservice_id = 'webservice{:03d}'.format(self._webservice_no)
                            webservice['id'] = webservice_id
                            self._webservice_no += 1

                        self.log.info(
                            "Order Transport {transport_id} to start Web Service {webservice_id}",
                            transport_id=hlid(transport_id),
                            webservice_id=hlid(webservice_id),
                            path=hluserid(path),
                        )

                        yield self._controller.call('crossbar.worker.{}.start_web_transport_service'.format(worker_id),
                                                    transport_id,
                                                    path,
                                                    webservice,
                                                    options=CallOptions())
                        self.log.info(
                            "Ok, Transport {transport_id} has started Web Service {webservice_id}",
                            transport_id=hlid(transport_id),
                            webservice_id=hlid(webservice_id),
                            path=hluserid(path),
                        )

            self.log.info(
                "Ok, {worker_logname} has started Transport {transport_id}",
                worker_logname=worker_logname,
                transport_id=hlid(transport_id),
            )



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from bitstring import BitStream

from crossbar.bridge.mqtt.protocol import (
    Connect,
    ConnACK,
    Subscribe,
    SubACK,
    Unsubscribe,
    UnsubACK,
    Publish,
    PubACK,
)

from twisted.trial.unittest import TestCase


class ConnectTests(TestCase):
    """
    Tests for Connect.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        # Header for CONNECT
        header = b"\x10\x13"
        # CONNECT without header, valid, client ID is "test123", clean session
        good = b"\x00\x04MQTT\x04\x02\x00\x00\x00\x07test123"

        event = Connect.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class ConnectAckTests(TestCase):
    """
    Tests for ConnectAck.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\x20\x02"
        good = b"\x00\x00"
        event = ConnACK.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class SubscribeTests(TestCase):
    """
    Tests for Subscribe.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\x82\x10"
        good = (b"\x00\x01\x00\x0b\x66\x6f\x6f\x2f\x62\x61\x72\x2f\x62\x61" b"\x7a\x00")
        event = Subscribe.deserialise((False, False, True, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class SubACKTests(TestCase):
    """
    Tests for SubACK.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\x90\x03"
        good = b"\x00\x01\x00"
        event = SubACK.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class PublishTests(TestCase):
    """
    Tests for Publish.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        # DUP0, QoS 0, Retain 0
        header = b"\x30\x1a"
        good = (b"\x00\x0b\x66\x6f\x6f\x2f\x62\x61\x72\x2f\x62\x61\x7a\x68\x65"
                b"\x6c\x6c\x6f\x20\x66\x72\x69\x65\x6e\x64\x73")

        event = Publish.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)

    def test_round_trip_qos1(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        # DUP0, QoS 1, Retain 0
        header = b"\x32\x1c"
        good = (b"\x00\x0b\x66\x6f\x6f\x2f\x62\x61\x72\x2f\x62\x61\x7a\x00\x02"
                b"\x68\x65\x6c\x6c\x6f\x20\x66\x72\x69\x65\x6e\x64\x73")

        event = Publish.deserialise((False, False, True, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class PubACKTests(TestCase):
    """
    Tests for PubACK.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\x40\x02"
        good = b"\x00\x02"

        event = PubACK.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class UnsubscribeTests(TestCase):
    """
    Tests for Unsubscribe.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\xa2\x19"
        good = (b"\x00\x03\x00\x15\x63\x6f\x6d\x2e\x65\x78\x61\x6d\x70\x6c\x65"
                b"\x2e\x6f\x6e\x63\x6f\x75\x6e\x74\x65\x72")

        event = Unsubscribe.deserialise((False, False, True, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)


class UnsubACKTests(TestCase):
    """
    Tests for UnsubACK.
    """
    def test_round_trip(self):
        """
        Deserialising a message and serialising it again results in the same
        binary message.
        """
        header = b"\xb0\x02"
        good = b"\x00\x03"

        event = UnsubACK.deserialise((False, False, False, False), BitStream(bytes=good))
        self.assertEqual(event.serialise(), header + good)



# coding=utf8

##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import re
import inspect
import uuid

import click

from autobahn.wamp import CallDetails

_ENVPAT_STR = r'^\$\{(.+)\}$'
_ENVPAT = re.compile(_ENVPAT_STR)


def hl(text, bold=False, color='yellow'):
    if not isinstance(text, str):
        text = '{}'.format(text)
    return click.style(text, fg=color, bold=bold)


def _qn(obj):
    if inspect.isclass(obj) or inspect.isfunction(obj) or inspect.ismethod(obj):
        qn = '{}.{}'.format(obj.__module__, obj.__qualname__)
    else:
        qn = 'unknown'
    return qn


def hltype(obj):
    qn = _qn(obj).split('.')
    text = hl(qn[0], color='yellow', bold=True) + hl('.' + '.'.join(qn[1:]), color='white', bold=True)
    return '<' + text + '>'


def hlid(oid):
    return hl('{}'.format(oid), color='blue', bold=True)


def hluserid(oid):
    if not isinstance(oid, str):
        oid = '{}'.format(oid)
    return hl('"{}"'.format(oid), color='yellow', bold=True)


def hlval(val, color='white'):
    return hl('{}'.format(val), color=color, bold=True)


def hlcontract(oid):
    if not isinstance(oid, str):
        oid = '{}'.format(oid)
    return hl('<{}>'.format(oid), color='magenta', bold=True)


def extract_member_oid(details: CallDetails) -> uuid.UUID:
    """
    Extract the XBR network member ID from the WAMP session authid (eg ``member_oid==72de3e0c-ca62-452f-8f09-2d3d30d1b511`` from ``authid=="member-72de3e0c-ca62-452f-8f09-2d3d30d1b511"``

    :param details: Call details.

    :return: Extracted XBR network member ID.
    """
    if details and details.caller_authrole == 'member' and details.caller_authid:
        return uuid.UUID(details.caller_authid[7:])
    else:
        raise RuntimeError('no XBR member identification in call details\n{}'.format(details))


def alternative_username(username):
    max_i = None
    for i in range(len(username) - 1, -1, -1):
        if username[i:].isdigit():
            max_i = i
    if max_i is not None:
        next = int(username[max_i:]) + 1
        alt_username = '{}{}'.format(username[:max_i], next)
    else:
        alt_username = '{}{}'.format(username, 1)
    return alt_username


def maybe_from_env(value):
    if isinstance(value, str):
        match = _ENVPAT.match(value)
        if match and match.groups():
            var = match.groups()[0]
            if var in os.environ:
                new_value = os.environ[var]
                return True, new_value
            else:
                print(
                    'WARNING: environment variable "{}" not set, but needed in XBR backend configuration'.format(var))
                return False, None
        else:
            return False, value
    else:
        return False, value
